*******************************************************************************
* Program.....: Import.prg                                                    *
* Authors.....: Joseph D. Booth                                               *
* Copyright...: 1998 Joseph D. Booth, All rights reserved.                    *
* Application.: CLIPWKS++                                                     *
* Version.....: 1.1                                                           *
* Compile.....: Xpp import -l -m -n -w                                        *
*               Limport(cWks_file,cDbf_file,cRange,cDeftype)                  *
*                                                                             *
*******************************************************************************

#include "CLIPWKS.CH"
#include "FILEIO.CH"

#define  MISSING_SPREADSHEET        - 1
#define  COULD_NOT_OPEN_FILE        - 2
#define  INVALID_RANGE              - 3
#define  INDETERMINATE_FIELD_TYPE   - 4
#define  DBF_CREATE_PROBLEM         - 5

#define  CELL_CONTENTS              arr_[ 1 ]
#define  CELL_TYPE                  arr_[ 2 ]
#define  CELL_ROW                   arr_[ 5 ]
#define  CELL_COLUMN                arr_[ 6 ]

#define  DEFAULT_TYPE               "N"


* Function:   Limport()
* Purpose:    To import a spreadsheet into a .DBF file
* Syntax:     Limport( cSpreadsheet,cDBF_file [,cRange] [,cDefType] )
*
* Arguments:  cSpreadsheet - name of the spreadsheet to read
*             cDBF_file    - name of .DBF file to import into
*             cRange       - optional range of cells to import
*             cDefType     - optional default field type
*
* Notes:      If the <cDBF_file> exists, it's structure is used
*             and data is appended into the file.  If you want
*             a new file created, be sure to ZAP the file before
*             Limport().
*
*             If the <cDBF_file> does not exist, then Limport()
*             uses the first row in the range of cells to
*             determine the types and sizes of the fields in
*             the .DBF file.  Be sure to specify a cell range
*             that does not include headers, since this would
*             result in all fields being character type, which
*             might not be what you want.  For example:
*
*                    A                 B                C
*             1     Name            Hire date        Salary
*             2   George Bush       01-JAN-89        120,000.00
*             3   Dan Quayle        01-JAN-89         80,000.00
*
*             Specify range A2..C3 rather than A1..C3.
*
* Returns:    nStatus      0 = Ok
*                         -1 = Missing spreadsheet
*                         -2 = Problem opening the file
*                         -3 = Invalid range specified
*                         -4 = Field type unknown in header
*                         -5 = Problem creating the .DBF file
*
******************************************************************************

FUNCTION limport( cWks_file, cDbf_file, cRange, cDefType )

   LOCAL x , aWks , flist := { }, arr_
   LOCAL cc, tmp, nSize, nDec, nFirstRec := 0, cType, ok := .T.

   IF ValType( cWks_file ) <> "C"          // Missing spreadsheet name, return
      RETURN MISSING_SPREADSHEET         // an error since we cannot process
   ENDIF                                 // the function call.

   IF ValType( cDbf_file ) <> "C"          // If the .DBF file is not specified
      x := At( ".", cWks_file )             // we will use the same file name as
      x := IF( x = 0, Len( cWks_file ), x - 1 )    // as the spreadsheet file

      cDbf_file := SubStr( cWks_file, 1, x ) + ".DBF"
   ENDIF

   IF cDefType == NIL                    // If you do not specify a cell type,
      cDefType := DEFAULT_TYPE           // the default will be numeric
   ENDIF


   aWks := Lread( cWks_file )            // Attempt to open for reading

   IF ! Empty( aWks )                      // If opened ok,
   *
      IF cRange == NIL
         cRange := Lr( WKS_TOP_ROW + 1, WKS_TOP_COL + 1, WKS_BOTTOM_ROW, WKS_BOTTOM_COL )
      ENDIF
      arr_ := Lgetrange( aWks, cRange )   // Get the select range of cells into an
      IF Empty( arr_ )                    // array.  Range can be a named range or
         Lclose( aWks )                   // two cells marking the top and bottom.
         RETURN INVALID_RANGE
      ENDIF
   *                                 // If the file does not exist, we need
      IF ! File( cDbf_file )               // to create it using the first row as
         FOR x := 1 TO Len( arr_[ 1 ] )     // the template structure.
            tmp     := arr_[ 1, x ]
            cType   := IF( tmp == NIL, cDefType, ValType( tmp ) )
            nSize   := 1
            nDec    := 0
            IF cType == "D"
               nSize := 8
            ELSEIF cType == "N"
               nDec  := 2
               nSize := 12
            ELSEIF cType == "C"
               nSize := Asc( SubStr( WKS_CELL_WIDTHS, x, 1 ) )
            ENDIF
            IF x < 27
               AAdd( flist, { "COL_" + Chr( 64 + x ), cType, nSize, nDec } )
            ELSE
               cc := Int( x / 26 )
               AAdd( flist, { "COL_" + Chr( 64 + cc ) + Chr( 65 + ( x % 26 ) ), cType, nSize, nDec } )
            ENDIF
         NEXT

         IF Len( flist ) > 0

            DBCreate( cDbf_file, flist )
            use ( cDbf_file ) Alias clipwks exclusive new
            FOR x := 1 TO Len( arr_ )
               append blank
            NEXT
         ELSE
            RETURN DBF_CREATE_PROBLEM
         ENDIF
      ELSE
         use ( cDbf_file ) Alias clipwks exclusive new
         nFirstRec := RecCount()
         FOR x := 1 TO Len( arr_ )
            append blank
         NEXT
         flist := DBStruct()
      ENDIF
      //
      // At this point, the array has been retrieved and the .DBF structure is
      // known, so we place elements into the appropiate fields and records.
      //
      IF ok
         FOR x := 1 TO Len( arr_ )
            goto nFirstRec + x
            FOR cc := 1 TO Len( arr_[ x ] )
               //
               // If the data returned from the spreadsheet agrees with the field
               // type, just place it directly into the .DBF file.
               //
               IF flist[ cc, 2 ] == ValType( arr_[ x, cc ] )
                  FieldPut( cc, arr_[ x, cc ] )
               ELSE
                  //
                  // Otherwise, call MakeType() to convert the spreadsheet data
                  // to the proper field type.
                  //
                  tmp := MakeType( flist[ cc, 2 ], arr_[ x, cc ] )
                  IF tmp <> NIL
                     FieldPut( cc, tmp )
                  ENDIF
               ENDIF
            NEXT
         NEXT
         use
      ENDIF
   *
      Lclose( aWks )
   ELSE
      RETURN COULD_NOT_OPEN_FILE
   ENDIF
   RETURN 0
*******************************************************************************

//   Function:  MakeType
//    Purpose:  This function attempts to convert the spreadsheet data into
//              the type of data the .DBF file is expecting in this field
//
//
//
//     Thanks:  John F. Kaster
////////////////////////////////////////////////////////////
STATIC FUNCTION MakeType( cToType, cField )
   LOCAL cFromType := ValType( cField )
   IF cToType == NIL .OR. cField == NIL
      RETURN NIL
   ENDIF

   IF cToType $ 'CM'                // Convert "From" type field to character
      IF cFromType == 'N'
         RETURN AllTrim( Str( cField ) )
      ELSEIF cFromType == 'L'
         RETURN IIf( cField, 'Y', 'N' )
      ELSEIF cFromType == 'D'
         RETURN DToC( cField )
      ELSE
         RETURN cField
      ENDIF
   ELSEIF cToType == 'N'            // Convert "From" type field to numeric
      IF cFromType $ 'CM'
         RETURN Val( cField )
      ELSEIF cFromType == 'L'
         RETURN IIf( cField, 1, 0 )
      ELSEIF cFromType == 'D'
         RETURN cField - CToD( "12/31/1899" )
      ELSE
         RETURN cField
      ENDIF
   ELSEIF cToType == 'D'            // Convert "From" type field to date
      IF cFromType $ 'CM'
         RETURN IF( CToD( cField ) <> "  /  /  ", CToD( cField ), NIL )
      ELSEIF cFromType == 'L' // Ain't no conversion
         RETURN NIL
      ELSEIF cFromType == 'N'
         RETURN  CToD( "12/31/1899" ) + cField
      ELSE
         RETURN cField
      ENDIF
   ELSEIF cToType == 'L'            // Convert "From" type field to logical
      IF cFromType $ 'CM'
         RETURN SubStr( Upper( cField ), 1, 1 ) $ 'YT'
      ELSEIF cFromType == 'D' // Ain't no conversion
         RETURN NIL
      ELSEIF cFromType == 'N'
         RETURN cField <>  0
      ELSE
         RETURN cField
      ENDIF
   ENDIF
   RETURN NIL
*******************************************************************************
* End of Limport *
*******************************************************************************
