*  Program.....: Miscwks.prg
*  Authors.....: Joseph D. Booth
*  Copyright...: 1998 Joseph D. Booth, All rights reserved.
*  Application.: CLIPWKS++
*  Version.....: 1.1
*  Compile.....: xpp miscwks /l /m /n /w
*  Functions...: Lalign(aWks,cAlignment)
*                Lcalcmode(aWks,cMethod)
*                Lcalcorder(aWks,cMethod)
*                Lfirstcell(aWks,cCell)
*                Lformat(aWks,cRange,cMethod)
*                Literate(aWks,nCount)
*                Lname(aWks,cName,cRange)
*                Lprotect(aWks,cMethod)
*                Lrange(aWks,nRows,nColumns)
*
***************************************************************************
#include "CLIPWKS.CH"


* Function:   Lalign()
* Purpose:    Set label alignment, (L)eft,(c)enter,(r)ight
* Syntax:     Lalign( aSpreadsheet,cAlignment )
* Arguments:  aSpreadsheet - Spreadsheet array handle
*             cAlignment   - Alignment method
* Notes:      Text in a cell may be justified left, right, or
*             centered.  The Lalign() function is used to set
*             the global alignment to used for text being
*             written to cells.
* Returns:    nStatus    - 0  All ok
*                        - 6  Invalid parameters
*****************************
FUNCTION lalign( aWks, cMethod )

   LOCAL retval := ERR_INVALID_PARAMETERS
   DEFAULT cMethod TO "L"

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ValType( cMethod ) == "C"

      WKS_ALIGNMENT := SubStr( cMethod, 1, 1 )
      retval        := ALL_OK

   ENDIF
   RETURN retval
*****************************************************************************
* End of Lalign *
*****************************************************************************


* Function:   Lcalcmode()
* Purpose:    Set calculation mode
* Syntax:     Lcalcmode( aSpreadsheet,cMode )
* Arguments:  aSpreadsheet - Spreadsheet handle array
*             cMode        - Calculation mode
* Notes:      The calculation mode determines how often a
*             spreadsheet's formula are recomputed.  In you
*             choice (A)utomatic, the spreadsheet will update
*             all formulas as soon as a cell value affecting the
*             formula is changed.  If you choice (M)anual, you
*             need to press a key, (usually F9), to recalculate
*             the spreadsheet.  (B)ackground allows the
*             spreadsheet to be recalculated while it is waiting
*             for keystrokes from the end-user.  If you create a
*             spreadsheet using CLIPWKS and set the calcmode to
*             (M)anual, be sure to have your end-user press F9
*             as soon they read in the spreadsheet.  CLIPWKS
*             does not calculate the values from the formulas it
*             writes to the spreadsheet.
* Returns:    nStatus
* See Also:   Lcalcorder()
********************************************
FUNCTION lcalcmode( aWks, cMethod )

   LOCAL retval := ERR_INVALID_PARAMETERS
   DEFAULT cMethod TO "A"

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ValType( cMethod ) == "C"

      WKS_CALCMODE := SubStr( cMethod, 1, 1 )
      retval       := ALL_OK

   ENDIF
   RETURN retval
*****************************************************************************
* End of Lcalcmode *
*****************************************************************************



* Function:   Lcalcorder()
* Purpose:    Set calculation order
* Syntax:     Lcalcorder( aSpreadsheet,cOrder )
* Arguments:  aSpreadsheet - Array of spreadsheet parameters
*             cOrder       - Calculation order,(R)owwise,(C)olumn,(N)atural
* Notes:      Calculation order determines when dependent cells
*             are recalculated in order to resolve any formulas.
*             Natural calculation order means that before a
*             formula is computed, each cell it refers to is
*             calculated first to ensure accurary.  In column
*             recalculation order, calculation starts in the
*             first column and proceeds down.  It ignores
*             formulas in other columns.  Rowwise works
*             similarly, calculating rows at a time.  The
*             preferred method is (N)atural.
* Returns:    nStatus
* See Also:   Lcalcmode()
*********************************
FUNCTION lcalcorder( aWks, cMethod )

   LOCAL retval := ERR_INVALID_PARAMETERS
   DEFAULT cMethod TO "N"                 // Default to (N)atural

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ValType( cMethod ) == "C"

      WKS_CALCORDER := SubStr( cMethod, 1, 1 )
      retval        := ALL_OK

   ENDIF
   RETURN retval
*****************************************************************************
* End of Lcalcorder *
*****************************************************************************

* Function:   Lfirstcell()
* Purpose:    Set first cell address
* Syntax:     Lfirstcell( aSpreadsheet,cAddress )
* Arguments:  aSpreadsheet - Spreadsheet array handle
*             cAddress     - First cell address
* Notes:      Setting the first cell allows you to have an automatic
*             macro startup when the user invokes your spreadsheet.
* Returns:    nStatus    - 0  All ok
*                        - 6  Invalid parameters
*****************************
FUNCTION lfirstcell( aWks, cCell )

   LOCAL retval := ERR_INVALID_PARAMETERS
   DEFAULT cCell TO lc( 1, 1 )

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ;
      ValType( cCell ) == "C" .AND. .NOT. Empty( cCell )

      WKS_FIRST_CELL  := cCell
      retval          := ALL_OK

   ENDIF
   RETURN retval
*****************************************************************************
* End of Lfirstcell *
*****************************************************************************

* Function:   Lformat()
* Purpose:    Set the global date format
* Syntax:     Lformat( aSpreadsheet, cRange,cFormat )
* Arguments:  aSpreadsheet - Spreadsheet array handle
*             cRange       - Range of cells
*             cFormat      - Format to set range to
*
* Returns:    nStatus
******************************
FUNCTION lformat( aWks, cRange, cFormat )

   LOCAL retval   := ERR_INVALID_PARAMETERS
   LOCAL r1, r2, c1, c2, tmp, jj, x, tt, nFormat, s1, s2
   LOCAL lProtect := ( WKS_PROTECTION == "P" )

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks )

      IF WKS_VERSION $ "L3|L4"                   // Lotus 3.x only

         IF SubStr( cFormat, 1, 1 ) $ "+-"        // If protection was specified
            lProtect := SubStr( cFormat, 1, 1 ) == "+"
            cFormat  := SubStr( cFormat, 2 )
         ENDIF

         nFormat := lcFormat( cFormat, aWks, lProtect )   // Break cRange into rows/column
         tmp     := Lmakerange( cRange, WKS_VERSION )

         IF ! Empty( tmp )
            r1  := Bin2W( SubStr( tmp, 1, 2 ) ) + 1
            c1  := Asc( SubStr( tmp, 4, 1 ) )
            s1  := Asc( SubStr( tmp, 3, 1 ) )
            r2  := Bin2W( SubStr( tmp, 5, 2 ) ) + 1
            c2  := Asc( SubStr( tmp, 8, 1 ) )
            s2  := Asc( SubStr( tmp, 7, 1 ) )
            FOR jj = r1 TO r2
               x   := AScan( WKS_FORMATS, { | z | Val( SubStr( z, 1, 6 ) ) = s1 * 10000 + jj } )
               IF x = 0
                  AAdd( WKS_FORMATS, Str( s1 * 10000 + jj, 6 ) + Replicate( Chr( 224 ), 256 ) )
                  x := Len( WKS_FORMATS )
               ENDIF
               FOR tt = c1 TO c2
                  WKS_FORMATS[ x ] := Stuff( WKS_FORMATS[ x ], 7 + tt, 1, Chr( nFormat ) )
               NEXT
            NEXT
            retval := ALL_OK
         ELSE
            retval := ERR_INVALID_RANGE
         ENDIF
      ENDIF
   ENDIF
   RETURN retval
*****************************************************************************
* End of Lformat *
*****************************************************************************

* Function:   Literate()
* Purpose:    Set the formula recalculation iteration flag
* Syntax:     Literate( aSpreadsheet, nCount )
* Arguments:  aSpreadsheet - Array of spreadsheet parameters
*             nCount       - Iteration count
* Notes:      On some complex formulas, multiple iterations are
*             needed to get an acceptable degree of accuracy.
*             You may specify the number of iterations to
*             perform on formula calculation using the
*             Literate() function.  The default value is one
*             iteration.
* Returns:    nStatus
* See Also:   Lcalcmode() Lcalcorder()
******************************
FUNCTION literate( aWks, nCount )

   LOCAL retval := ERR_INVALID_PARAMETERS
   DEFAULT nCount TO 1

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ValType( nCount ) == "N"

      WKS_ITERATION := nCount
      retval        := ALL_OK

   ENDIF
   RETURN retval
*****************************************************************************
* End of Literate *
*****************************************************************************


* Function:   Lname()
* Purpose:    Write a named range into the spreadsheet
* Syntax:     Lname( aSpreadsheet, cName, cRange )
* Arguments:  aSpreadsheet - Array of spreadsheet parameters
*             cName        - Name for specified range
*             cRange       - Range of cells
* Notes:      A spreadsheet allows a range of cells to have a
*             name assigned to them.  This assists in
*             documenting function calls.  The Lname() function
*             writes an range name out to the spreadsheet.  If a
*             formula references the range, it will display the
*             range name when the formula is view.
* Returns:    nStatus
*********************************
FUNCTION lname( aWks, cName, cRange )

   LOCAL retval := ERR_INVALID_PARAMETERS, _rngexp

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ;
      ValType( cName ) == "C" .AND. ValType( cRange ) == "C"

      _rngexp := Lmakerange( cRange, WKS_VERSION )

      IF ! Empty( _rngexp ) .AND. ! Empty( cName )
         AAdd( WKS_NAMED_RANGES, { cName, _rngexp } )
         retval := ALL_OK
      ELSEIF Empty( _rngexp )
         retval := ERR_INVALID_RANGE
      ELSE
         retval := ERR_MISSING_PARAMETER
      ENDIF

   ENDIF
   RETURN retval
*****************************************************************************
* End of Lname *
*****************************************************************************


* Function:   Lprotect()
* Purpose:    Set the global protection flag in the spreadsheet
* Syntax:     Lprotect( aSpreadsheet,cSetting | lSetting )
* Arguments:  aSpreadsheet - Array of spreadsheet parameters
*             cSetting     - ON  | P  - Enable global protection
*                            OFF | U  - Disable global protection
*             lSetting     - TRUE - Enable  FALSE - disable
* Notes:      Lprotect is used to enable or disable global
*             protection.  When protection is enabled, protected
*             cells may not be modified.
* Returns:    nStatus
********************************************
FUNCTION lprotect( aWks, cMethod )

   LOCAL retval := ERR_INVALID_PARAMETERS
   DEFAULT cMethod TO "P"

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ValType( cMethod ) $ "CL"

      cMethod        := IF( ValType( cMethod ) == "L", IF( cMethod, "P", "U" ), cMethod )
      WKS_PROTECTION := SubStr( cMethod, 1, 1 )
      retval         := ALL_OK

   ENDIF
   RETURN retval
*****************************************************************************
* End of Lprotect *
*****************************************************************************


* Function:   Lrange()
* Purpose:    Write number of rows/columns into spreadsheet
* Syntax:     Lrange( aSpreadsheet,nRows,nColumns )
* Arguments:  aSpreadsheet - Array of spreadsheet parameters
*             nRows        - Number of rows in spreadsheet
*             nColumns     - Number of columns in spreadsheet
* Notes:      The Lrange() command is used to specify how many
*             active rows and columns the spreadsheet has.
* Returns:    nStatus
********************************************
FUNCTION lrange( aWks, nRows, nCols )

   LOCAL retval := ERR_INVALID_PARAMETERS
   DEFAULT nRows TO 1
   DEFAULT nCols TO 1

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ;
      ValType( nRows ) == "N" .AND. ValType( nCols ) == "N"

      WKS_ROWS    := nRows
      WKS_COLUMNS := nCols
      retval      := ALL_OK

   ENDIF
   RETURN retval
*****************************************************************************
* End of Lrange *
*****************************************************************************
