***************************************************************
* Program.......: READWKS.PRG
* Authors.......: Joseph D. Booth
* Copyright.....: 1998 Joseph D. Booth, All rights reserved.
* Application...: CLIPWKS++
* Version.......: 1.1
* Compile.......: Xpp readwks -l -m
*                 Lfindfirst( aWks [,cColumn] )  - Get first cell
*                 Lfindnext( aWks )              - Get next cell
*                 Lread(cFilename,lUpdate)       - Read a spreadsheet
*
* Special Thanks: Ira Emus
*                 Rob Hannah
*                 John Kaster
*
****************************************************************************
#include "CLIPWKS.CH"
#include "FILEIO.CH"

* Function:   Lfindfirst()
* Purpose:    To return the first cell in the spreadsheet
* Syntax:     Lfindfirst( aWks [,xColumn] )
*
* Arguments:  aWks      - Spreadsheet handle array
*             xColumn   - Column letter to find first cell of
*             nColumn   - Column number to find first cell of
* Returns:    aCell     - Array of cell information
*
* Notes:      Lfindfirst() positions the spreadsheet to the first cell
*             in the spreadsheet or in the specified column.  It also
*             returns an array of cell information.  Subsequent calls
*             to Lfindnext() will obtain the next cell.
* See Also:   Lfindnext()  Lgetcell()
**********************************************

FUNCTION lfindfirst( aWks, xColumn )
   LOCAL retval , ln

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks )
      //
      // No column specified, go to first cell in spreadsheet
      //
      IF xColumn == NIL

         // Fill buffer,pointer,offset

         aWks[ 10 ]   := { Lfileread( WKS_HANDLE, WKS_DATA_OFFSET, BUFFSIZE ), ;
                           1, WKS_DATA_OFFSET }
         ln         := Bin2W( SubStr( WKS_READ_BUFFER, 3, 2 ) )
         retval     := limget( aWks, SubStr( WKS_READ_BUFFER, 1, ln + 4 ) )

         WKS_READ_POINTER := ln + 5

      ELSEIF SubStr( WKS_VERSION, 1, 1 ) == "E" .AND. ValType( xColumn ) == "N"

      ELSE

         IF ValType( xColumn ) == "C"
            xColumn := IF( Len( xColumn ) = 1, Asc( xColumn ) - 65, ;
                           ( Asc( xColumn ) - 64 ) * 26 + ( Asc( SubStr( xColumn, 2, 1 ) ) - 65 ) )
         ENDIF
         ln     := Bin2W( SubStr( WKS_CPI, xColumn * 4 + 1, 2 ) )
         retval := Lgetcell( aWks, lc( ln + 1, xColumn + 1 ) )
      ENDIF
   ENDIF
   RETURN retval
*****************************************************************************



* Function:   Lfindnext()
* Purpose:    To return subsequent cells from either Lfindfirst() or Lget()
* Syntax:     Lfindnext( aWks )
*
* Arguments:  aWks      - Spreadsheet handle array
* Returns:    aCell     - Array of cell information
*
* Notes:      Lfindnext() reads the next cell in the spreadsheet. It
*             may start as a result of either a call to Lfindfirst()
*             or to Lgetcell().  If a cell is found, an array of cell
*             information is returned.  If no cell is found, an empty
*             array is returned.
* See Also:   Lfindfirst()  Lgetcell()
**********************************************
FUNCTION lfindnext( aWks )
   LOCAL retval := { } , buf , ln, op //, cTmp, x
   IF ValType( aWks ) == "A" .AND. ! Empty( aWks )

      WHILE .T.

         IF WKS_READ_POINTER > ( BUFFSIZE - 5 )           // Not enough room in buffer\

            WKS_READ_OFFSET += WKS_READ_POINTER        // to determine length
            buf := Space( BUFFSIZE )                   // memory
            FRead( WKS_HANDLE, @buf, BUFFSIZE )
            WKS_READ_BUFFER  := SubStr( WKS_READ_BUFFER, WKS_READ_POINTER ) + buf
            WKS_READ_POINTER := 1
         ENDIF

         op       := Bin2W( SubStr( WKS_READ_BUFFER, WKS_READ_POINTER, 2 ) )
         ln       := Bin2W( SubStr( WKS_READ_BUFFER, WKS_READ_POINTER + 2, 2 ) )

         IF WKS_READ_POINTER + 4 + ln > Len( WKS_READ_BUFFER )

            WKS_READ_OFFSET += WKS_READ_POINTER
            buf := Space( BUFFSIZE )
            FRead( WKS_HANDLE, @buf, BUFFSIZE )
            WKS_READ_BUFFER  := SubStr( WKS_READ_BUFFER, WKS_READ_POINTER ) + buf
            WKS_READ_POINTER := 1

         ENDIF
         retval     := limget( aWks, SubStr( WKS_READ_BUFFER, WKS_READ_POINTER, ln + 4 ) )
         WKS_READ_POINTER += ln + 4
         IF SubStr( WKS_VERSION, 1, 1 ) == "E"
            IF op == 10 .AND. ln == 0
               retval := { }
               EXIT
            ENDIF
            IF ! Empty( retval ) .AND. ;
               retval[ 1 ] = NIL .AND. ! Empty( WKS_READ_BUFFER )
               LOOP
            ENDIF
         ENDIF
         EXIT
      ENDDO
   ENDIF
   RETURN retval

*****************************************************************************


* Function:   Lgetrange()
* Purpose:    To return an array of spreadsheet cells
* Syntax:     Lgetrange( aWks,cRange )
*
* Arguments:  aWks      - Spreadsheet handle array
*             cRange    - Range specifier or range name
* Returns:    aContents - Contents of all cells within range
*
**********************************************


FUNCTION lgetrange( aWks, cRange )
   LOCAL arr_ := { }, aCell := { }, nTrow, nTcol, nBrow, nBcol , cTmp , jj , x
   LOCAL xx, yy, nSheet
   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ! Empty( cRange )
      IF At( "..", cRange ) == 0      // Looking for a named range
         //
         FOR jj = 1 TO Len( WKS_RANGES )
            IF Trim( cRange ) == Trim( WKS_RANGES[ jj, 1 ] )
               cRange := WKS_RANGES[ jj, 2 ]
               jj     := 99999
            ENDIF
         NEXT
      ENDIF
      IF ( x := At( "..", cRange ) ) <> 0                      // Found a range to extract
         // Determine number of cells with range
         cTmp      := Lmakerange( cRange, WKS_VERSION )
         IF ! Empty( cTmp )
            IF WKS_VERSION == "L3" .OR. WKS_VERSION == "L4"
               nTrow := Bin2W( SubStr( cTmp, 1, 2 ) )
               nTcol := Asc( SubStr( cTmp, 4, 1 ) )
               nSheet := Asc( SubStr( cTmp, 3, 1 ) )
               nBrow := Bin2W( SubStr( cTmp, 5, 2 ) )
               nBcol := Asc( SubStr( cTmp, 8, 1 ) )
            ELSEIF SubStr( WKS_VERSION, 1, 1 ) == "E"
               nSheet := - 1
               nTrow := Bin2W( SubStr( cTmp, 1, 2 ) )
               nTcol := Bin2W( SubStr( cTmp, 3, 2 ) )
               nBrow := Bin2W( SubStr( cTmp, 5, 2 ) )
               nBcol := Bin2W( SubStr( cTmp, 7, 2 ) )
            ELSE
               nSheet := - 1
               nTrow := Bin2W( SubStr( cTmp, 3, 2 ) )
               nTcol := Bin2W( SubStr( cTmp, 1, 2 ) )
               nBrow := Bin2W( SubStr( cTmp, 7, 2 ) )
               nBcol := Bin2W( SubStr( cTmp, 5, 2 ) )
            ENDIF
            arr_      := Array( nBrow - nTrow + 1, nBcol - nTcol + 1 )
            IF WKS_VERSION <> "QP"
               FOR xx = nTrow + 1 TO nBrow + 1
                  FOR yy = nTcol + 1 TO nBcol + 1
                     aCell     := Lgetcell( aWks, lc( xx, yy, nSheet + 1 ) )
                     IF ! Empty( aCell )
                        xx = nBrow + 3
                        yy = nBcol + 3
                     ENDIF
                  NEXT
               NEXT
            ELSE
               FOR xx = nTcol + 1 TO nBcol + 1
                  FOR yy = nTrow + 1 TO nBrow + 1
                     aCell     := Lgetcell( aWks, lc( yy, xx, nSheet ) )
                     IF ! Empty( aCell )
                        xx = nBcol + 3
                        yy = nBrow + 3
                     ENDIF
                  NEXT
               NEXT
            ENDIF
            WHILE ! Empty( aCell )
               IF aCell[ 5 ] > ( nBrow + 1 ) .AND. aCell[ 6 ] > ( nBcol + 1 )
                  EXIT
               ENDIF
               aCell[ 5 ] -= nTrow
               aCell[ 6 ] -= nTcol
               IF aCell[ 5 ] <= Len( arr_ ) .AND. aCell[ 6 ] <= Len( arr_[ 1 ] ) ;
                  .AND. aCell[ 5 ] > 0 .AND. aCell[ 6 ] > 0  .AND. arr_[ aCell[ 5 ], aCell[ 6 ] ] == NIL
                  arr_[ aCell[ 5 ], aCell[ 6 ] ] := aCell[ 1 ]
               ENDIF
               aCell := Lfindnext( aWks )
            ENDDO
         ENDIF
      ENDIF
   ENDIF
   RETURN arr_
*****************************************************************************




* Function:   Lreplace()
* Purpose:    To replace an existing cell's contents
* Syntax:     Lreplace( aWks,cCell,xData )
* Arguments:  aWks      - Spreadsheet handle array
*             cCell     - Cell address to change
* Returns:    nStatus   - Status of operation
*
**********************************************
FUNCTION lreplace( aWks, xData, cCell )
   LOCAL retval  := - 6, arr_ := { }, tmp, x //, cell_spot
   LOCAL rewrite := .F.
   LOCAL cHold
   IF ValType( aWks ) == "A" .AND. ! Empty( aWks )
      IF aWks[ 4 ]
         arr_ := IF( ValType( cCell ) == "A" , arr_, Lgetcell( aWks, cCell ) )
         IF ! Empty( arr_ )
            IF arr_[ 2 ] == ValType( xData )
               tmp := lfileread( WKS_HANDLE, arr_[ 7 ], 512 )
               DO CASE
************** Character Data ************************************************
               CASE ( arr_[ 2 ] == "C" .AND. WKS_VERSION == "QP" ) .AND. Len( xData ) <= arr_[ 3 ]
                  tmp := Stuff( tmp, 12, arr_[ 3 ], PadR( xData, arr_[ 3 ] ) )
               CASE ( arr_[ 2 ] == "C" .AND. WKS_VERSION $ "L3|L4" ) .AND. Len( xData ) <= arr_[ 3 ]
                  tmp := Stuff( tmp, 10, arr_[ 3 ], Lstring( xData, arr_[ 3 ] ) )
               CASE arr_[ 2 ] == "C" .AND. WKS_VERSION $ "E2|E3|E4" .AND. Len( xData ) <= arr_[ 3 ]
                  tmp := Stuff( tmp, 13, arr_[ 3 ], PadR( xData, arr_[ 3 ] ) )
               CASE arr_[ 2 ] == "C" .AND. Len( xData ) <= arr_[ 3 ]
                  tmp := Stuff( tmp, 11, arr_[ 3 ], Lstring( xData, arr_[ 3 ] ) )
               CASE arr_[ 2 ] == "C"
                  rewrite := .T.
************** Numeric / Date ************************************************
               CASE arr_[ 2 ] $ "DN" .AND. WKS_VERSION $ "L3|L4"
                  x := IF( arr_[ 2 ] == "D", Abs( xData - CToD( "12/30/1899" ) ), xData )
                  IF arr_[ 1 ] <> x      // Don't replace same value
                     IF arr_[ 3 ] == 2
                        IF x >= - 16384 .AND. x <= 16383 .AND. x == Int( x )
                           tmp := Stuff( tmp, 9, 2, shortnum( x ) )
                        ELSE
                           rewrite := .T.
                        ENDIF
                     ELSEIF arr_[ 3 ] == 10
                        tmp := Stuff( tmp, 9, 10, treal( x ) )
                     ENDIF
                  ENDIF
               CASE arr_[ 2 ] $ "DN"
                  IF arr_[ 2 ] == "D" .AND. arr_[ 3 ] == 2
                     IF xData - CToD( "12/30/1899" ) > 32767
                        rewrite := .T.
                     ENDIF
                  ENDIF
                  x := IF( arr_[ 2 ] == "D", Abs( xData - CToD( "12/30/1899" ) ), xData )
                  DO CASE
                  CASE arr_[ 3 ] == 2
                     IF x >= - 32767 .AND. x <= 32768 .AND. x == Int( x )
                        tmp := Stuff( tmp, 10, 2, I2Bin( x ) )
                     ELSE
                        rewrite := .T.
                     ENDIF
                  CASE arr_[ 3 ] == 8
                     tmp := Stuff( tmp, 10, 8, ieereal( x ) )
                  CASE arr_[ 3 ] == 4 .AND. Left( WKS_VERSION, 1 ) == "E"
                     cHold := ieereal( xData )
                     cHold := SubStr( cHold, 5, 4 )
                     bit( cHold, 8, .F. )
                     bit( cHold, 7, .F. )
                     tmp   := Stuff( tmp, 11, 4, cHold )
                  ENDCASE
************** Logical *******************************************************
               CASE arr_[ 2 ] == "L" .AND. WKS_VERSION == "QU"
                  IF arr_[ 1 ] <> xData     // Don't replace if same value
                     IF xData             // .T.
                        tmp := Stuff( tmp, 10, 8, ieereal( 1 ) )
                        tmp := Stuff( tmp, 20, 1, Chr( 114 ) )
                     ELSE
                        tmp := Stuff( tmp, 10, 8, ieereal( 0 ) )
                        tmp := Stuff( tmp, 20, 1, Chr( 113 ) )
                     ENDIF
                  ENDIF
               CASE arr_[ 2 ] == "L" .AND. WKS_VERSION $ "L3|L4"
                  IF arr_[ 1 ] <> xData     // Don't replace if same value
                     IF xData             // .T.
                        tmp := Stuff( tmp, 9, 10, treal( 1 ) )
                        tmp := Stuff( tmp, 19, 1, Chr( 52 ) )
                     ELSE
                        tmp := Stuff( tmp, 9, 10, treal( 0 ) )
                        tmp := Stuff( tmp, 19, 1, Chr( 51 ) )
                     ENDIF
                  ENDIF
               CASE arr_[ 2 ] == "L"
                  x := IF( WKS_VERSION == "QP", 24, 20 )
                  IF arr_[ 1 ] <> xData     // Don't replace if same value
                     IF xData             // .T.
                        tmp := Stuff( tmp, 10, 8, ieereal( 1 ) )
                        tmp := Stuff( tmp, x, 1, Chr( 52 ) )
                     ELSE
                        tmp := Stuff( tmp, 10, 8, ieereal( 0 ) )
                        tmp := Stuff( tmp, x, 1, Chr( 51 ) )
                     ENDIF
                  ENDIF
               ENDCASE
               IF ! rewrite
                  FSeek( WKS_HANDLE, arr_[ 7 ], FS_SET )
                  retval := IF( FWrite( WKS_HANDLE, tmp, Len( tmp ) ) == Len( tmp ), 0, - 2 )
               ENDIF
            ELSE
               rewrite := .T.
            ENDIF
         ELSE
            rewrite := .T.
         ENDIF
         IF rewrite
            FSeek( WKS_HANDLE, - 4, FS_END )
            tmp := Lput( aWks, xData, cCell,, .F. )
            IF Left( WKS_VERSION, 1 ) <> "E"
               FWrite( WKS_HANDLE, tmp + EOF_OPCODE )
            ELSE
               FWrite( WKS_HANDLE, tmp + E2_EOF_OPCODE )
            ENDIF
            aWks[ 5, 2 ] := aWks[ 5, 2 ] + Len( tmp )
            retval := 0
         ENDIF
      ELSE
         retval := - 12
      ENDIF
   ENDIF
   RETURN retval
*****************************************************************************




* Function:   Lrows()
* Purpose:    To determine number of rows in spreadsheet
* Syntax:     Lrows( aWks )
* Arguments:  aWks      - Spreadsheet handle array
* Returns:    Number of rows
*             -6       - Invalid spreadsheet array
* See Also:   Lcols()
**********************************************


FUNCTION lrows( aWks )
   LOCAL retval := - 6
   IF ValType( aWks ) == "A" .AND. ! Empty( aWks )
      retval := WKS_BOTTOM_ROW + 1      // Rows in spreadsheet
   ENDIF
   RETURN retval
*****************************************************************************


* Function:   Lcols()
* Purpose:    To determine number of columns in spreadsheet
* Syntax:     Lcols( aWks )
* Arguments:  aWks      - Spreadsheet handle array
* Returns:    Number of columns
*             -6       - Invalid spreadsheet array
* See Also:   LRows()
**********************************************
FUNCTION lcols( aWks )
   LOCAL retval := - 6
   IF ValType( aWks ) == "A" .AND. ! Empty( aWks )
      retval := WKS_BOTTOM_COL + 1      // Columns in spreadsheet
   ENDIF
   RETURN retval
*****************************************************************************



* Function:   Lread()
* Purpose:    To open a spreadsheet for reading
* Syntax:     Lread( cFilename,lUpdate,lIndex,lDates )
* Arguments:  cFilename - Name of spreadsheet to read
*             lUpdate   - Whether CLIPWKS can update the spreadsheet
* Returns:    NIL
**********************************************
FUNCTION lread( cFilename, LUpdate, lDates )

   LOCAL aWks := { }, nHandle, cVersion, opcode := 0, nLength := 0, cTmp := "", k, nRow
   LOCAL tmp := "", y, subtype, cSubform, cPattern, cFormat, colptr, do_it := .F., jj, xTmp
   LOCAL nWhere := 0,  x, z
   //LOCAL ok := .T.
   DEFAULT LUpdate  TO  .F.
   DEFAULT lDates   TO  .T.
   nHandle  := FOpen( cFilename, IF( LUpdate, FO_READWRITE, FO_READ ) )
   cVersion := Lversion( nHandle )               // Make sure it is a Lotus/Quattro file
   IF ! Empty( cVersion )
      aWks := { cFilename, ;                   // Spreadsheet file name
                cVersion, ;                    // Version
                nHandle, ;                     // Handle file opened on
                LUpdate, ;                     // Allowed to update?
                { 0, 0 }, ;                       // Data offset/end
                { 0, 0, 0, 0, 9 }, ;                  // Coordinates, global width
                "", ;                          // CPI information
                { }, ;                          // Date bitmap for Lotus 3.x
                { }, ;                          // Ranges
                { "", 0, 0 }, ;                    // Buffer, pointer, offset
                "", ;                          // Global format for Lotus 3.x
                Replicate( Chr( 9 ), 256 ),  ;      // Cell width's
                { }, { }, 0 }                      // Excel extra elements

      IF SubStr( cVersion, 1, 1 ) == "E"
         EXCEL_DATE_FORMAT := CToD( "12/30/1899" )     // Date offset
      ENDIF

      WKS_FILE_SIZE := FSeek( nHandle, 0, FS_END ) // Get file size
      FSeek( nHandle, 0, FS_SET )                  // Top of file
      bread(, nHandle )                          // Initialize the buffer
      WHILE ! lWksEof( opcode, cVersion )
         opcode := Bin2W( bread( 2, nHandle ) )      // Read operation code
         nLength := Bin2W( bread( 2, nHandle ) )      // and data length
         IF nLength > 0                        // If data is present, then
            cTmp := bread( nLength, nHandle )     // extract the data
         ENDIF
         IF cVersion == "L3" .OR. cVersion == "L4"
            DO CASE
               // Check for start of data section
            CASE opcode = 0
               WKS_TOP_ROW      := Bin2W( SubStr( cTmp, 5, 2 ) )
               WKS_TOP_COL      := Asc( SubStr( cTmp, 8, 1 ) )
               WKS_BOTTOM_ROW   := Bin2W( SubStr( cTmp, 9, 2 ) )
               WKS_BOTTOM_COL   := Asc( SubStr( cTmp, 12, 1 ) )
            CASE opcode = 6
               WKS_GLOBAL_WIDTH := Asc( SubStr( cTmp, 5, 1 ) )
               aWks[ 12 ]         := StrTran( aWks[ 12 ], Chr( 9 ), Chr( WKS_GLOBAL_WIDTH ) )
            CASE opcode >= 20 .AND. opcode <= 26
               WKS_DATA_OFFSET  := bread(, - 1 ) - nLength - 5       // Compute offset in file
               FSeek( nHandle, WKS_DATA_OFFSET, FS_SET )            // Postion the file and
               EXIT
            CASE opcode == 31
               IF Empty( WKS_CPI )
                  tmp     := I2Bin( 0 ) + I2Bin( 0 )
                  WKS_CPI := Replicate( tmp, 256 )
               ENDIF
               IF Asc( cTmp ) == 0
                  y := Asc( SubStr( cTmp, 2, 1 ) )
                  WKS_CPI := Stuff( WKS_CPI, y * 4 + 1, 4, ;
                                    SubStr( cTmp, 5, 2 ) + SubStr( cTmp, - 2, 2 ) )
               ENDIF
            CASE opcode = 19 .AND. lDates
               subtype := Asc( SubStr( cTmp, 2, 1 ) )
               DO CASE
               CASE subtype == 0
                  colptr  := 0
                  k       := Bin2W( SubStr( cTmp, 3, 2 ) )
                  cFormat := SubStr( cTmp, 5, nLength - 4 )
                  cPattern := Str( k + 1, 6 ) + Replicate( Chr( 0 ), 32 )
                  WHILE ! Empty( cFormat )
                     k := 1
                     IF bit( cFormat, 25 )
                        cTmp    := SubStr( cFormat, 1, 5 )
                        cFormat := SubStr( cFormat, 6 )
                        k       := Asc( SubStr( cTmp, 5, 1 ) ) + 1
                     ELSE
                        cTmp    := SubStr( cFormat, 1, 4 )
                        cFormat := SubStr( cFormat, 5 )
                     ENDIF
                     // See if cell is a date
                     IF bit( cTmp, 2 ) .AND. bit( cTmp, 3 ) .AND. bit( cTmp, 4 )
                        cSubform := IF( bit( cTmp, 5 ), "1", "0" ) + ;
                                    IF( bit( cTmp, 6 ), "1", "0" ) + ;
                                    IF( bit( cTmp, 7 ), "1", "0" ) + ;
                                    IF( bit( cTmp, 8 ), "1", "0" )
                        IF cSubform == "0010" .OR. cSubform == "0011" .OR. ;
                        cSubform == "0100" .OR. cSubform == "1001" .OR. ;
                        cSubform == "1010"
                        FOR jj = 1 TO k
                           colptr ++
                           bit( cPattern, colptr + 48, .T. )
                        NEXT
                        do_it := .T.
                     ELSE
                        colptr += k
                     ENDIF
                  ELSE
                     colptr += k
                  ENDIF
               ENDDO
               IF do_it
                  AAdd( WKS_DATE_BITMAP, cPattern )
               ENDIF

            CASE subtype == 1
               WK3_GLOBAL_FORMAT := Asc( SubStr( cTmp, 5, 1 ) )

            CASE subtype == 2
               nRow := Bin2W( SubStr( cTmp, 3, 2 ) ) + 1
               k := Bin2W( SubStr( cTmp, 7, 2 ) ) + 1
               jj := AScan( WKS_DATE_BITMAP, { | y | Val( SubStr( y, 1, 6 ) ) == k } )
               IF jj > 0
                  cPattern := Str( nRow, 6 ) + SubStr( aWks[ 8, jj ], 7, 32 )
                  AAdd( WKS_DATE_BITMAP, cPattern )
               ENDIF

            ENDCASE
         CASE opcode == 9                         // Named ranges
            IF Bin2W( SubStr( cTmp, 1, 2 ) ) == 0
               y := SubStr( cTmp, 3, 16 )
               k := At( Chr( 0 ), y )
               IF k > 0
                  y := PadR( SubStr( y, 1, k - 1 ), 16 )
               ENDIF
               xTmp := lc( Bin2W( SubStr( cTmp, 19, 2 ) ) + 1, Asc( SubStr( cTmp, 22, 1 ) ) + 1, ;
                           Asc( SubStr( cTmp, 21, 1 ) ) + 1 ) + ".."
               xTmp += lc( Bin2W( SubStr( cTmp, 23, 2 ) ) + 1, ;
                           Asc( SubStr( cTmp, 26, 1 ) ) + 1, Asc( SubStr( cTmp, 25, 1 ) ) + 1 )
               AAdd( aWks[ 9 ], { y, xTmp } )
            ENDIF
         CASE opcode == 7       // Cell width
            IF Asc( SubStr( cTmp, 1, 1 ) ) == 0
               FOR k = 5 TO Len( cTmp ) STEP 2
                  y := Asc( SubStr( cTmp, k, 1 ) )
                  aWks[ 12 ] := Stuff( aWks[ 12 ], y + 1, 1, SubStr( cTmp, k + 1, 1 ) )
               NEXT
            ENDIF
         ENDCASE

      ELSEIF SubStr( cVersion, 1, 1 ) == "E"

         DO CASE
         CASE opcode == 30 .OR. opcode == 1055
            AAdd( EXCEL_FORMATS, IF( opcode == 30, SubStr( cTmp, 2 ), SubStr( cTmp, 4 ) ) )
         CASE opcode = 0 .OR. opcode = 512
            WKS_TOP_ROW       := Bin2W( SubStr( cTmp, 1, 2 ) )
            WKS_TOP_COL       := Bin2W( SubStr( cTmp, 5, 2 ) )
            WKS_BOTTOM_ROW    := Bin2W( SubStr( cTmp, 3, 2 ) )
            WKS_BOTTOM_COL    := Bin2W( SubStr( cTmp, 7, 2 ) )
         CASE opcode == 11 .OR. opcode == 523
            x := IF( opcode == 11, 9, 13 )
            EXCEL_FIRST_ROW := Bin2W( SubStr( cTmp, 5, 2 ) )
            FOR y := x TO Len( cTmp ) STEP 4
               AAdd( EXCEL_INDEX_RECORD, Bin2L( SubStr( cTmp, y, 4 ) ) )
            NEXT

         CASE ( opcode == 8 .OR. opcode == 520 ) .AND. WKS_DATA_OFFSET < 1
            WKS_DATA_OFFSET := nWhere + 20 + IF( opcode == 8, Bin2W( SubStr( cTmp, 12, 2 ) ), Bin2W( SubStr( cTmp, 11, 2 ) ) )
            x  := Len( EXCEL_INDEX_RECORD )
            y  := EXCEL_INDEX_RECORD[ x ]
            FSeek( nHandle, y, FS_SET )                  // Next data cell
            bread(, nHandle )                          // Initialize the buffer
            LOOP

         CASE opcode == 34 .AND. Bin2W( cTmp ) == 1
            EXCEL_DATE_FORMAT := CToD( "01/01/1904" )     // Date offset

         CASE opcode = 67
            AAdd( EXCEL_XF_RECORDS, cTmp )
         CASE opcode = 579
            AAdd( EXCEL_XF_RECORDS, cTmp )
         CASE opcode = 1091
            AAdd( EXCEL_XF_RECORDS, cTmp )
         CASE opcode == 85
            WKS_GLOBAL_WIDTH := Bin2W( SubStr( cTmp, 1, 2 ) )
            aWks[ 12 ]         := StrTran( aWks[ 12 ], Chr( 9 ), Chr( WKS_GLOBAL_WIDTH ) )
         CASE opcode == 36
            y                := Asc( SubStr( cTmp, 1, 1 ) )
            z                := Asc( SubStr( cTmp, 2, 1 ) )
            k                := Int( Bin2W( SubStr( cTmp, 3, 2 ) ) / 256 )
            FOR x := y TO z
               aWks[ 12 ] := Stuff( aWks[ 12 ], y + 1, 1, Chr( k ) )
            NEXT
         CASE opcode == 125
            y                := Bin2W( SubStr( cTmp, 1, 2 ) )
            z                := Bin2W( SubStr( cTmp, 3, 2 ) )
            k                := Int( Bin2W( SubStr( cTmp, 5, 2 ) ) / 256 )
            FOR x := y TO z
               aWks[ 12 ] := Stuff( aWks[ 12 ], y + 1, 1, Chr( k ) )
            NEXT
         CASE opcode == 24
         CASE opcode == 536       // Range name
            x := Asc( SubStr( cTmp, 4, 1 ) )
            y := Bin2W( SubStr( cTmp, 5, 2 ) )
            IF Asc( SubStr( cTmp, 7 + x, 1 ) ) == 45 .AND. x > 1       // Found a range
               xTmp := lr( Bin2W( SubStr( cTmp, 8 + x, 2 ) ) + 1, ;
                           Asc( SubStr( cTmp, 12 + x, 1 ) ) + 1,  ;
                           Bin2W( SubStr( cTmp, 10 + x, 2 ) ) + 1, ;
                           Asc( SubStr( cTmp, 13 + x, 1 ) ) + 1 )
               AAdd( aWks[ 9 ], { SubStr( cTmp, 7, x ), xTmp } )
            ENDIF
         ENDCASE
      ELSE
         DO CASE
            // Check for start of data section
         CASE opcode >= 12 .AND. opcode <= 16
            WKS_DATA_OFFSET := bread(, - 1 ) - nLength - 5    // Compute offset in file
            FSeek( nHandle, WKS_DATA_OFFSET, FS_SET )        // Postion the file and
            EXIT
         CASE opcode == 6                        // Range record
            WKS_TOP_ROW      := Bin2W( SubStr( cTmp, 3, 2 ) )
            WKS_TOP_COL      := Bin2W( SubStr( cTmp, 1, 2 ) )
            WKS_BOTTOM_ROW   := Bin2W( SubStr( cTmp, 7, 2 ) )
            WKS_BOTTOM_COL   := Bin2W( SubStr( cTmp, 5, 2 ) )
         CASE opcode == 11 .OR. opcode == 71      // Named ranges
            IF WKS_VERSION == "QP"
               y := SubStr( cTmp, 2, Asc( cTmp ) )
            ELSE
               y := SubStr( cTmp, 1, 16 )
            ENDIF
            y := PadR( StrTran( y, Chr( 0 ), " " ), 16 )
            AAdd( aWks[ 9 ], { y, lr( Bin2W( SubStr( cTmp, 19, 2 ) ) + 1, ;
                                      Bin2W( SubStr( cTmp, 17, 2 ) ) + 1, ;
                                      Bin2W( SubStr( cTmp, 23, 2 ) ) + 1, ;
                                      Bin2W( SubStr( cTmp, 21, 2 ) ) + 1 ) } )
         CASE opcode == 150
            tmp := I2Bin( 0 ) + I2Bin( 0 )
            WKS_CPI := Replicate( tmp, 256 )
            FOR k = 1 TO Len( cTmp ) STEP 6
               y := Bin2W( SubStr( cTmp, k, 2 ) )
               WKS_CPI := Stuff( WKS_CPI, y * 4 + 1, 4, SubStr( cTmp, k + 2, 4 ) )
            NEXT
         CASE opcode == 8       // Cell width
            y := Bin2W( SubStr( cTmp, 1, 2 ) )
            aWks[ 12 ] := Stuff( aWks[ 12 ], y + 1, 1, SubStr( cTmp, 3, 1 ) )
         CASE opcode == 7       // Global width
            WKS_GLOBAL_WIDTH := Bin2W( SubStr( cTmp, 7, 2 ) )
            aWks[ 12 ]         := StrTran( aWks[ 12 ], Chr( 9 ), Chr( WKS_GLOBAL_WIDTH ) )
         ENDCASE
      ENDIF
      nWhere += ( 4 + nLength )
   ENDDO
ENDIF
RETURN aWks
*****************************************************************************
* End of Lread *
*****************************************************************************



STATIC FUNCTION lWksEof( nOpcode, cVer )
   IF SubStr( cVer, 1, 1 ) == "E"
      RETURN nOpcode == 10
   ENDIF
   RETURN nOpcode == 1
*****************************************************************************
* End of Lwkseof() *
*****************************************************************************

* Purpose:    To return the contents of a cell
* Syntax:     Lget( aSpreadsheet,cCell )
* Arguments:  aSpreadsheet    - Spreadsheet handle array
*             cCell           - Cell address to look for
* Returns:    xCell_contents  - "" if cell not found
* See Also:   Ltype()
**********************************************
FUNCTION lget( aWks, cCell )
   LOCAL arr_ := Lgetcell( aWks, cCell )
   RETURN IF( Empty( arr_ ), "", arr_[ 1 ] )
*****************************************************************************
* End of Lgetcell() *
*****************************************************************************
* Purpose:    To return the type of a cell
* Syntax:     Ltype( aSpreadsheet,cCell )
* Arguments:  aSpreadsheet    - Spreadsheet handle array
*             cCell           - Cell address to look for
* Returns:    xCell_type      - "" if cell not found
*                                C - Character
*                                D - Date
*                                L - Logical
*                                N - Numeric
*                                F - Formula
* See Also:   Lget()
**********************************************
FUNCTION ltype( aWks, cCell )
   LOCAL arr_ := Lgetcell( aWks, cCell )
   RETURN IF( Empty( arr_ ), "", arr_[ 2 ] )
*****************************************************************************
* End of Lgetcell() *
*****************************************************************************
* Purpose:    To return an array of cell attributes
* Syntax:     Lgetcell( aSpreadsheet,cCell )
* Arguments:  aSpreadsheet    - Spreadsheet handle array
*             cCell           - Cell address to look for
* Returns:    aCell_info      - 1 = Contents
*                               2 = Cell type
*                               3 = Cell length
*                               4 = Cell address
*                               5 = Row number
*                               6 = Column number
*                               7 = File offset
* See Also:   Lget() Ltype()
**********************************************
FUNCTION lgetcell( aWks, cCell )
   LOCAL cCoords := Lcell( cCell, WKS_VERSION ), nRow, nCol, nSheet //, cRetval := NIL
   LOCAL nStart, nAvg, ok := .T., arr_ := { }, nEnd := aWks[ 5, 2 ], tmp, k, x, op, ln, nEarly, cAddr
   LOCAL nFirstCol := 0, nLastCol := - 1, nWhere, cTmp
   LOCAL nSize

   IF SubStr( WKS_VERSION, 1, 1 ) == "E"
      nRow   := Bin2W( SubStr( cCoords, 1, 2 ) )             // Row
      nCol   := Bin2W( SubStr( cCoords, 3, 2 ) )             // Column
   ELSEIF WKS_VERSION == "L3" .OR. WKS_VERSION == "L4"
      nRow   := Bin2W( SubStr( cCoords, 1, 2 ) )             // Row
      nSheet := Asc( SubStr( cCoords, 3, 1 ) )               // Spreadsheet
      nCol   := Asc( SubStr( cCoords, 4, 1 ) )               // Column
   ELSE
      nRow   := Bin2W( SubStr( cCoords, 3, 2 ) )             // Row
      nCol   := Bin2W( SubStr( cCoords, 1, 2 ) )             // Column
   ENDIF

   IF SubStr( WKS_VERSION, 1, 1 ) == "E"                   // Special treatment for Excel
      //
      // Determine where the row records are stored, by taking the row number
      // and looking it up in the EXCEL_INDEX_RECORD.  Go to this offset and
      // start reading ROW records until you find one where
      //
      IF ! Empty( EXCEL_INDEX_RECORD )
         x      := Int( ( nRow - EXCEL_FIRST_ROW ) / 32 ) + 1
         nWhere := EXCEL_INDEX_RECORD[ x ]
         FSeek( WKS_HANDLE, nWhere, FS_SET )
         bread(, WKS_HANDLE )
         op := Bin2W( bread( 2, WKS_HANDLE ) )
         ln := Bin2W( bread( 2, WKS_HANDLE ) )
         x := 0
         nStart := nWhere + 20
         WHILE ( op == 8 .AND. ln == 18 ) .OR. ( op == 520 .AND. ln = 16 ) .AND. nLastCol < 0
            cTmp := bread( ln, WKS_HANDLE )
            x         := IF( op == 8, Bin2W( SubStr( cTmp, 12, 2 ) ), Bin2W( SubStr( cTmp, 11, 2 ) ) )
            nStart    += x
            IF Bin2W( SubStr( cTmp, 1, 2 ) ) == nRow
               nFirstCol := Bin2W( SubStr( cTmp, 3, 2 ) )
               nLastCol  := Bin2W( SubStr( cTmp, 5, 2 ) )
            ENDIF
            nWhere += ( 4 + ln )
            op := Bin2W( bread( 2, WKS_HANDLE ) )
            ln := Bin2W( bread( 2, WKS_HANDLE ) )
            IF nLastCol > 0 .AND. ( op == 8 .OR. op == 520 )
               cTmp := bread( ln, WKS_HANDLE )
               x         := IF( op == 8, Bin2W( SubStr( cTmp, 12, 2 ) ), Bin2W( SubStr( cTmp, 11, 2 ) ) )
               nEnd := nStart + x - 1
            ENDIF
         ENDDO
      ELSE
         nStart    := WKS_DATA_OFFSET
         nEnd      := aWks[ 5, 2 ]
         nFirstCol := WKS_TOP_COL
         nLastCol  := WKS_BOTTOM_COL
      ENDIF

      IF nCol >= nFirstCol .AND. nCol <= nLastCol
         x := 0
         WHILE Empty( x )
            x := llocate( aWks, cCoords, nStart, nEnd )
            IF x > 0
               // Read previous few bytes to see if legitimate cell
               // If ok, exit the loop
               x -= 4
               x   := Max( x, WKS_DATA_OFFSET )
               tmp := lfileread( WKS_HANDLE, x, BUFFSIZE )
               op  := Bin2W( SubStr( tmp, 1, 2 ) )
               ln  := Bin2W( SubStr( tmp, 3, 2 ) )
               IF op > 4096
                  tmp := SubStr( tmp, 3 )
                  op  := Bin2W( SubStr( tmp, 1, 2 ) )
                  ln  := Bin2W( SubStr( tmp, 3, 2 ) )
               ENDIF
               IF Bin2W( SubStr( tmp, 5, 2 ) ) <> nRow .OR. ;
                  Bin2W( SubStr( tmp, 7, 2 ) ) <> nCol .OR. ;
                  ! is_cell( WKS_VERSION, op, ln )
                  nStart := x + 9
                  x      := 0
                  IF nStart >= nEnd        // Cell was not found
                     RETURN arr_
                  ENDIF
                  LOOP
               ELSE
                  EXIT
               ENDIF
            ENDIF
         ENDDO
         //
         IF x > 0
            WKS_READ_BUFFER   := tmp
            WKS_READ_POINTER  := 1
            WKS_READ_OFFSET   := x
            arr_              := Limget( aWks, SubStr( WKS_READ_BUFFER, 1, ln + 4 ) )
            WKS_READ_POINTER  := ln + 5
         ENDIF
         //
      ENDIF
      RETURN arr_
   ENDIF


   IF nRow >= WKS_TOP_ROW .AND. nRow <= WKS_BOTTOM_ROW     // Is requested cell
      IF nCol >= WKS_TOP_COL .AND. nCol <= WKS_BOTTOM_COL  // within spreadsheet range?

         IF Len( WKS_CPI ) == 1024 .AND. nSheet == 0
            ok := nRow >= Bin2W( SubStr( WKS_CPI, nCol * 4 + 1, 2 ) )  .AND. ;
                  nRow <= Bin2W( SubStr( WKS_CPI, nCol * 4 + 3, 2 ) )

         ENDIF

         IF ok           // Cell should be in the spreadsheet
            // See if the cell is in the buffer
            IF ! Empty( WKS_READ_BUFFER )
               k     := 1
               nSize := Len( WKS_READ_BUFFER )
               WHILE k < nSize
                  op    := Bin2W( SubStr( WKS_READ_BUFFER, k, 2 ) )
                  ln    := Bin2W( SubStr( WKS_READ_BUFFER, k + 2, 2 ) )
                  IF k + 2 + ln > nSize
                     EXIT
                  ENDIF
                  cAddr := SubStr( WKS_READ_BUFFER, k + 4, 4 )
                  IF cAddr == cCoords .AND. is_cell( WKS_VERSION, op, ln )
                     WKS_READ_POINTER := k
                     RETURN Lfindnext( aWks )
                  ENDIF
                  k += ( ln + 4 )
                  IF k + 4 > nSize    // Not enough room in buffer to read op,ln
                     EXIT
                  ENDIF
               ENDDO
            ENDIF

            // Attempt to make a reasonable first guess as to where
            nAvg  := ( aWks[ 5, 2 ] - WKS_DATA_OFFSET ) / ( ( WKS_BOTTOM_COL - WKS_TOP_COL ) * ( WKS_BOTTOM_ROW - WKS_TOP_ROW ) )
            nStart := Max( nRow * nCol * nAvg + WKS_DATA_OFFSET - 512, WKS_DATA_OFFSET )
            nEarly := nStart
            k     := 0

            WHILE k == 0
               k := llocate( aWks, cCoords, nStart, nEnd ) // From current spot to END
               IF k > 0
                  // Read previous few bytes to see if legitimate cell
                  // If ok, exit the loop
                  IF ! ( WKS_VERSION $ "L3|L4" )
                     k -= 5
                     k   := Max( k, WKS_DATA_OFFSET )
                     tmp := lfileread( WKS_HANDLE, k, BUFFSIZE )
                     op  := Bin2W( SubStr( tmp, 1, 2 ) )
                     ln  := Bin2W( SubStr( tmp, 3, 2 ) )
                     IF .NOT. is_cell( WKS_VERSION, op, ln )
                        nStart := k + 9
                        k      := 0
                     ENDIF
                  ELSE
                     k -= 4
                     k   := Max( k, WKS_DATA_OFFSET )
                     tmp := lfileread( WKS_HANDLE, k, BUFFSIZE )
                     op  := Bin2W( SubStr( tmp, 1, 2 ) )
                     ln  := Bin2W( SubStr( tmp, 3, 2 ) )
                     IF .NOT. is_cell( WKS_VERSION, op, ln )
                        nStart := k + 9
                        k      := 0
                     ENDIF
                  ENDIF
               ENDIF
               IF k < 0
                  IF nStart == WKS_DATA_OFFSET .AND. nEnd < WKS_FILE_SIZE
                     EXIT
                  ENDIF
                  IF WKS_DATA_OFFSET < nStart
                     nEnd   := nEarly - 1
                     nStart := WKS_DATA_OFFSET
                     k      := 0
                  ENDIF
               ENDIF
            ENDDO
            IF k > 0          // Found the cell
               aWks[ 10 ] := { tmp, 1, k }
               arr_     := Limget( aWks, SubStr( WKS_READ_BUFFER, 1, ln + 4 ) )
               WKS_READ_POINTER := ln + 5
               IF  arr_[ 2 ] == "S"
                  arr_  := Lfindnext( aWks )
               ENDIF
            ENDIF
         ENDIF
      ENDIF
   ENDIF
   RETURN arr_
*****************************************************************************
* End of Lgetcell *
*****************************************************************************

STATIC FUNCTION Is_cell( cVers, op, ln )
   DO CASE
   CASE cVers == "L3" .OR. cVers == "L4"
      RETURN ( op = 20 .AND. ln == 4 ) .OR. ( op = 21 .AND. ln = 4 ) .OR. ;
      ( op = 22 .AND. ln < 519 ) .OR. ( op = 23 .AND. ln = 14 ) .OR. ;
      ( op = 24 .AND. ln == 6 ) .OR. ( op = 25 .AND. ln < 2049 ) .OR. ;
      ( op = 26 .AND. ln < 519 )
   CASE cVers == "E3" .OR. cVers == "E4"
      RETURN ( op == 638 .AND. ln = 10 ) .OR. ( op == 519 ) .OR. ( op == 1030 ) .OR. ;
      ( op == 513 .AND. ln == 6 ) .OR. ( op == 515 .AND. ln == 14 ) .OR. ;
      ( op == 516 ) .OR. ( op == 517 .AND. ln == 8 ) .OR. ( op == 518 )
   CASE cVers == "E2"
      RETURN ( op == 1 .AND. ln == 7 ) .OR. ( op == 2 .AND. ln == 9 ) .OR. ;
      ( op == 3 .AND. ln == 15 ) .OR. ( op == 4 ) .OR. ( op == 5 .AND. ln == 9 ) .OR. ;
      ( op == 6 ) .OR. ( op == 7 )
   OTHERWISE
      RETURN ( op == 12 .AND. ln = 5 )  .OR. ( op = 13 .AND. ln = 7 ) .OR. ;
      ( op == 14 .AND. ln = 13 ) .OR. ( op = 15 .AND. ln < 256 ) .OR. ;
      ( op == 16 .AND. ln < 2065 ) .OR. ( op = 51 .AND. ln < 256 )
   ENDCASE
   RETURN .F.
*****************************************************************************
* End of Is_cell *
*****************************************************************************

* Function:   Llocate()
* Purpose:    To locate a cell within the spreadsheet
* Syntax:     llocate( aSpreadsheet,cCell )
* Arguments:  aSpreadsheet - spreadsheet handle array
*             cCell        - Cell coordinates to locate
* Returns:    nOffset      - Offset in file where cell found
* Class:      Internal
*************************************************
STATIC FUNCTION llocate( aWks, cText, nStart, nEnd )
   LOCAL cBuffer  := Space( BUFFSIZE ), nPointer := 0
   LOCAL cOverlap := '', nOverlap := Len( cText )
   LOCAL nBytes   := BUFFSIZE
   //LOCAL nSavePos, nLen := 1, tmp
   cText     :=  cText

   FSeek( WKS_HANDLE, nStart, FS_SET )                        // Position file pointer
   WHILE nBytes == BUFFSIZE .AND. nPointer == 0
      nBytes  := FRead( WKS_HANDLE, @cBuffer, BUFFSIZE )
      IF nBytes <> BUFFSIZE                       // At last buffer
         EXIT
      ENDIF
      nPointer := At( cText, cOverlap + SubStr( cBuffer, 1, nBytes ) )
      IF nPointer == 0  // Text wasn't found yet - assign new cOverlap
         cOverlap := Right( SubStr( cBuffer, 1, nBytes ), nOverlap )
      ENDIF
      IF nPointer == 0 .AND. ( FSeek( WKS_HANDLE, 0, FS_RELATIVE ) > nEnd )
         EXIT
      ENDIF
   ENDDO

   IF nPointer == 0     // Look for text in last buffer
      nPointer := At( cText, cOverlap + SubStr( cBuffer, 1, nBytes ) )
   ENDIF

   IF nPointer > 0      // The text pointer was found
      nPointer -= ( Len( cOverlap ) + nBytes + 1 )  // get amount to move back
      IF FSeek( WKS_HANDLE, nPointer, 1 )  <= nEnd             // Move back
         RETURN FSeek( WKS_HANDLE, 0, 1 )
      ENDIF
   ENDIF
   RETURN - 1
*****************************************************************************
* End of Llocate() *
*****************************************************************************

* Function:   Lfileread()
* Purpose:    To read a portion of an open file
* Syntax:     lfileread( nHandle,nStart,nBytes )
* Arguments:  nHandle      - file handle to read from
*             nStart       - starting offset into the file
*             nBytes       - number of bytes to read
* Returns:    cData        - Data read from the file
* Class:      Internal
*************************************************
STATIC FUNCTION lfileread( nFh, nStart, nSize )
   LOCAL cBuf := Space( nSize ), nBytes
   FSeek( nFh, nStart )
   nBytes  := FRead( nFh, @cBuf, nSize )
   RETURN SubStr( cBuf, 1, nBytes )
*****************************************************************************
* End of Lfileread *
*****************************************************************************

*   Function:  Limget()
*    Purpose:  To return an array of cell information
*     Syntax:  limget( aSpreadsheet,cBuffer )
*  Arguments:  aSpreadsheet - spreadsheet handle array
*              cBuffer      - data from the current cell
*    Returns:  aCell_info   - array of cell information
*************************************************
STATIC FUNCTION limget( aWks, cBuffer )
   LOCAL retval     := { }, fmtbyte, fm, tst, is_spec, is_date := .F., nRow, nCol, cAddr, nSheet
   LOCAL nOper      := Bin2W( SubStr( cBuffer, 1, 2 ) )
   LOCAL nLength    := Bin2W( SubStr( cBuffer, 3, 2 ) )
   LOCAL xData      := SubStr( cBuffer, 5, nLength ), nAttr, nX, lInt, lDiv, cTmp, x, y, z
   LOCAL is_logical := .F.
//
// Test for Lotus 3.x
//
///////////////////////
   DO CASE
   CASE WKS_VERSION $ "L3|L4"
      nCol    := Asc( SubStr( xData, 4, 1 ) ) + 1
      nRow    := Bin2W( SubStr( xData, 1, 2 ) ) + 1
      nSheet  := Asc( SubStr( xData, 3, 1 ) ) + 1
      cAddr   := Lc( nRow, nCol, nSheet )

      //
      // Need to somehow check if a date is flagged
      //
      ///////////////////////////
      IF ! Empty( WKS_DATE_BITMAP )
         fm   := AScan( WKS_DATE_BITMAP[ 1 ], nRow )
         IF fm > 0
            is_date := bit( WKS_DATE_BITMAP[ 2, fm ], nCol )
         ENDIF
      ENDIF
      DO CASE
      CASE nOper == 20                        // Error cell
         retval := { "ERR", "E", 3 }
      CASE nOper == 21                        // Not applicable cell
         retval := { "N/A", "NA", 3 }
      CASE nOper == 22                        // Label
         retval := { StrTran( SubStr( xData, 6, nLength ), Chr( 0 ), Chr( 32 ) ), "C", nLength - 5 }
      CASE nOper == 23                        // Number cell
         retval :=  treal( SubStr( xData, 5, 10 ) )
         IF is_date
            retval := { CToD( "12/30/1899" ) + retval, "D", 10  }
         ELSE
            retval := { retval, "N", 10  }
         ENDIF
      CASE nOper == 24                        // Small number cell
         retval :=  shortnum( SubStr( xData, 5, 2 ) )
         IF is_date
            retval := { CToD( "12/30/1899" ) + retval, "D", 2  }
         ELSE
            retval := { retval, "N", 2  }
         ENDIF
      CASE nOper == 25                        // Formula cell
         retval := treal( SubStr( xData, 5, 10 ) )
         IF SubStr( xData, 15, 2 ) == Chr( 51 ) + Chr( 03 )
            retval := { .F., "L", 10 }
         ELSEIF SubStr( xData, 15, 2 ) == Chr( 52 ) + Chr( 03 )
            retval := { .T., "L", 10 }
         ELSEIF SubStr( xData, 15, 2 ) == Chr( 55 ) + Chr( 03 )
            retval := { CToD( "12/30/1899" ) + retval, "F", 10 }
         ELSEIF SubStr( xData, - 2, 2 ) == Chr( 54 ) + Chr( 03 )
            retval := { CToD( "12/30/1899" ) + retval, "F", 10 }
         ELSEIF is_date
            retval := { CToD( "12/30/1899" ) + retval, "F", 10 }
         ELSE
            retval := { treal( SubStr( xData, 5, 10 ) ), "F", 10 }
         ENDIF
      CASE nOper == 26
         retval := { StrTran( SubStr( xData, 5, nLength ), Chr( 0 ), Chr( 32 ) ), "C", nLength - 4 }
      ENDCASE
//
// All versions of Excel
//
//////////////////////////////////////////////////////////
   CASE WKS_VERSION == "E2" .OR. WKS_VERSION == "E3" .OR. WKS_VERSION == "E4"
      nCol    := Bin2W( SubStr( xData, 3, 2 ) ) + 1
      nRow    := Bin2W( SubStr( xData, 1, 2 ) ) + 1
      nAttr   := 1
      IF WKS_VERSION == "E2"
         nAttr := Asc( SubStr( xData, 6, 1 ) )
         IF nAttr > 128 ; nAttr -= 128 ; ENDIF
         IF nAttr > 64  ; nAttr -= 64  ; ENDIF
         nAttr ++
         is_date := nAttr >= 13 .AND. nAttr <= 16 .OR. nAttr == 21
      ELSE
         x       := Bin2W( SubStr( xData, 5, 2 ) )    // Pointer to extended format
         z       := ""
         IF x >= 0 .AND. x < Len( EXCEL_XF_RECORDS )
            y       := Asc( SubStr( EXCEL_XF_RECORDS[ x + 1 ], 2, 1 ) )
            IF y >= 0 .AND. y < Len( EXCEL_FORMATS )
               z       := Upper( EXCEL_FORMATS[ y + 1 ] )
            ENDIF
         ENDIF
         is_date := "D" $ z .AND. "/" $ z
      ENDIF

      cAddr   := Lc( nRow, nCol )
      DO CASE
      CASE nOper == 1 .OR. nOper == 513                // Blank cell
         retval := { "", "B", 0 }
      CASE nOper == 2                                  // Integer
         IF is_date
            retval := { EXCEL_DATE_FORMAT + Bin2W( SubStr( xData, 8, 2 ) ), "D", 2 }
         ELSE
            retval := { Bin2W( SubStr( xData, 8, 2 ) ), "N", 2 }
         ENDIF
      CASE nOper == 3                                  // Floating point number
         IF is_date
            retval := { EXCEL_DATE_FORMAT + real2iee( SubStr( xData, 8, 8 ) ), "D", 8 }
         ELSE
            retval := { real2iee( SubStr( xData, 8, 8 ) ), "N", 8 }
         ENDIF
      CASE nOper == 4                                  // Label (text) cell
         retval := { SubStr( xData, 9, Asc( SubStr( xData, 8, 1 ) ) ), "C", ;
                     Asc( SubStr( xData, 8, 1 ) ) }
      CASE nOper == 5
         IF Asc( SubStr( xData, 8, 1 ) ) == 0
            retval := { Asc( SubStr( xData, 9, 1 ) ) == 1, "L", 1 }
         ELSEIF Asc( SubStr( xData, 8, 1 ) ) == 1
            retval := { "ERR", "E", 1 }
         ENDIF
      CASE nOper == 6
         IF SubStr( xData, 14, 2 ) == HEX_FFFF   // Special value
            IF Asc( SubStr( xData, 8, 1 ) ) == 0             // String
            ELSEIF Asc( SubStr( xData, 8, 1 ) ) == 1         // Boolean
               retval := { Asc( SubStr( xData, 10, 1 ) ) = 1, "L", 8 }
            ELSEIF Asc( SubStr( xData, 8, 1 ) ) == 2         // Error
               retval := { Asc( SubStr( xData, 10, 1 ) ), "E", 8 }
            ENDIF
         ELSE
            IF is_date
               retval := { EXCEL_DATE_FORMAT + real2iee( SubStr( xData, 8, 8 ) ), "F", 8 }
            ELSE
               retval := { real2iee( SubStr( xData, 8, 8 ) ), "F", 8 }
            ENDIF
         ENDIF
      CASE nOper == 515                                // Floating point number
         IF is_date
            retval := { EXCEL_DATE_FORMAT + real2iee( SubStr( xData, 7, 8 ) ), "D", 8 }
         ELSE
            retval := { real2iee( SubStr( xData, 7, 8 ) ), "N", 8 }
         ENDIF
      CASE nOper == 516                                // Label (text) cell
         retval := { SubStr( xData, 9, Bin2W( SubStr( xData, 7, 2 ) ) ), "C", ;
                     Bin2W( SubStr( xData, 7, 2 ) ) }
      CASE nOper == 517
         IF Asc( SubStr( xData, 7, 1 ) ) == 0
            retval := { Asc( SubStr( xData, 8, 1 ) ) == 1, "L", 1 }
         ELSEIF Asc( SubStr( xData, 7, 1 ) ) == 1
            retval := { "ERR", "E", 1 }
         ENDIF
      CASE nOper == 518 .OR. nOper == 1030
         IF SubStr( xData, 13, 2 ) == HEX_FFFF   // Special value
            IF Asc( SubStr( xData, 7, 1 ) ) == 0             // String
            ELSEIF Asc( SubStr( xData, 7, 1 ) ) == 1         // Boolean
               retval := { Asc( SubStr( xData, 9, 1 ) ) = 1, "L", 8 }
            ELSEIF Asc( SubStr( xData, 7, 1 ) ) == 2         // Error
               retval := { Asc( SubStr( xData, 9, 1 ) ), "E", 8 }
            ENDIF
         ELSE
            IF is_date
               retval := { EXCEL_DATE_FORMAT + real2iee( SubStr( xData, 7, 8 ) ), "F", 8 }
            ELSE
               retval := { real2iee( SubStr( xData, 7, 8 ) ), "F", 8 }
            ENDIF
         ENDIF
      CASE nOper == 638                                // RK number
         cTmp := SubStr( xData, 7, 4 )
         lDiv := bit( cTmp, 8 )                           // Divide by 100
         lInt := bit( cTmp, 7 )                           // Integer bit
         bit( cTmp, 8, .F. )
         bit( cTmp, 7, .F. )
         nX   := IF( lInt, Bin2L( cTmp ), ;
                     real2iee( Chr( 0 ) + Chr( 0 ) + Chr( 0 ) + Chr( 0 ) + cTmp ) )
         IF lInt
            nX /= 4
         ENDIF
         IF lDiv
            nX /= 100
         ENDIF
         IF is_date
            retval := { EXCEL_DATE_FORMAT + nX, "D", 4 }
         ELSE
            retval := { nX, "N", 4 }
         ENDIF
      OTHERWISE
         retval := { NIL, NIL, NIL }
      ENDCASE
//
// Lotus 2.x, Quattro, and all versions of Quattro Pro
//
//////////////////////////////////////////////////////////
   OTHERWISE
      fmtbyte := SubStr( xData, 1, 1 )                // Format byte
      nCol    := Bin2W( SubStr( xData, 2, 2 ) ) + 1
      nRow    := Bin2W( SubStr( xData, 4, 2 ) ) + 1
      cAddr   := Lc( nRow, nCol )
      is_spec := ( bit( fmtbyte, 2 ) .AND. bit( fmtbyte, 3 ) .AND. bit( fmtbyte, 4 ) )
      IF is_spec
         tst     := IF( bit( fmtbyte, 6 ), 4, 0 ) + IF( bit( fmtbyte, 7 ), 2, 0 ) + IF( bit( fmtbyte, 8 ), 1, 0 )
         is_date := tst > 1 .AND. tst < 5   // We've got a date
      ENDIF
      DO CASE
      CASE nOper == 10
         retval := { }
      CASE nOper == 12                                 // Blank cell
         retval := { "", "B", 0 }
      CASE nOper == 13                                 // Integer cell
         IF is_date
            retval := { CToD( "12/30/1899" ) + Bin2I( SubStr( xData, 6, 2 ) ), "D", 2 }
         ELSE
            retval := { Bin2I( SubStr( xData, 6, 2 ) ), "N", 2 }
         ENDIF
      CASE nOper == 14                                 // Real number
         retval  := { real2iee( SubStr( xData, 6, 8 ) ), "N", 8 }
         IF is_date
            retval := { CToD( "12/30/1899" ) + real2iee( SubStr( xData, 6, 8 ) ), "D", 8 }
         ENDIF
      CASE nOper == 16                                 // Formula
         retval  := real2iee( SubStr( xData, 6, 8 ) )
         fm      := Asc( SubStr( xData, nLength - 1, 1 ) )
         IF Left( WKS_VERSION, 1 ) == "Q"
            IF ( fm == 113 .OR. fm == 114 )  // TRUE/FALSE
               retval := { ( retval == 1 ), "L", 8 }
               is_logical := .T.
            ELSEIF ( fm == 115 .OR. fm == 118 .OR. ;
                     fm == 206 .OR. fm = 142 )                   // Date formulas
               is_date := .T.
            ENDIF

         ELSE
            IF ( fm == 51 .OR. fm == 52 )  // TRUE/FALSE
               retval := { ( retval == 1 ), "L", 8 }
               is_logical := .T.
            ELSEIF ( fm == 55 .OR. fm == 54 )    // Date formulas
               is_date := .T.
            ENDIF
         ENDIF
         IF is_date
            retval := { CToD( "12/30/1899" ) + retval, "D", 8 }
         ELSEIF .NOT.  is_logical
            retval := { retval, "F", 8 }
         ENDIF
      CASE nOper == 15      // Label
         IF WKS_VERSION <> "QP"
            retval := SubStr( xData, 7, nLength )
            retval := { SubStr( retval, 1, Len( retval ) - 1 ), "C", nLength - 6 }
         ELSE
            tst    := Asc( SubStr( xData, 6, 1 ) )
            retval := { SubStr( xData, 8, tst ), "C", nLength - 7 }
         ENDIF
      CASE nOper == 51                // String value of formula
         IF WKS_VERSION <> "QP"
            retval := SubStr( xData, 6, nLength )
            retval := { SubStr( retval, 1, Len( retval ) - 1 ), "C", nLength - 5 }
         ELSE
            tst    := Asc( SubStr( xData, 6, 1 ) )
            retval := { SubStr( xData, 8, tst ), "C", nLength - 7 }
         ENDIF
      ENDCASE
   ENDCASE
//
// If we found a value, fill the array
//
/////////////////
   IF ! Empty( retval )
      AAdd( retval, cAddr )                                 // Cell address
      AAdd( retval, nRow )                                  // Row of spreadsheet
      AAdd( retval, nCol )                                  // Column
      AAdd( retval, WKS_READ_OFFSET + WKS_READ_POINTER - 1 )    // Physical file offset
   ENDIF
   RETURN retval
*****************************************************************************
* End of Limget *
*****************************************************************************










