*  Program.....: Writewks.prg
*  Authors.....: Joseph D. Booth
*  Copyright...: 1998 Joseph D. Booth, All rights reserved.
*  Application.: CLIPWKS++
*  Version.....: 1.1
*  Compile.....: xpp writewks /l /m /n /w
*                Lblank(aWks,cCell [,cFormat])
*                Lcreate(cFilename,cType)
*                Lfunc(aWks,cFormula,cCell [,cFormat])
*                Lput(aWks,xData,cCell [,cFormat])
*                Lwidth(aWks,cColumn|nColumn,nWidth)
*                Lparse()
*
****************************************************************************
#include "CLIPWKS.CH"

STATIC  nHighest_sheet  := 1


#define QP_FORMULAS     1
#define QP_CELLS        2
#define QP_RANGES       3

* Function:    Lblank()
* Purpose:     Write a blank cell into the worksheet
* Syntax:      Lblank( aSpreadsheet, cCell, cFormat )
* Arguments:   aSpreadsheet - spreadsheet handle array
*              cCell        - Cell address to write blank into
*              cFormat      - Format byte to write in cell
* Notes:       Lblank() allows you to write a blank cell into a
*              spreadsheet.  Unless the cell is referred to by a
*              formula or has a different format, blank cells are
*              normally not saved within a spreadsheet.
* Returns:     nStatus    - 0  All ok
*                         - 4  Invalid cell
*                         - 6  Invalid parameters
* See Also:    Lput()
***********************************
FUNCTION lblank( aWks, cCell, cFormat )

   LOCAL retval := ERR_INVALID_PARAMETERS, cell_spot, nFormat

   IF cFormat <> NIL
      nFormat := lcformat( cFormat, aWks )
   ELSE
      nFormat := DEFAULT_FORMAT
   ENDIF

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks ) .AND. ValType( cCell ) == "C"

      cell_spot := Lcell( cCell, WKS_VERSION )                // Cell location

      IF ! Empty( cell_spot )                                 // If a valid cell

         DO CASE
         CASE WKS_VERSION == "L3" .OR. WKS_VERSION == "L4"
         CASE WKS_VERSION == "E2"
            Bwrite( E2_BLANK + cell_spot + I2Bin( nFormat ) + Chr( 0 ), aWks )
         CASE WKS_VERSION == "E3" .OR. WKS_VERSION == "E4"
            Bwrite( E3_BLANK + cell_spot + I2Bin( nFormat ), aWks )
         OTHERWISE
            bwrite( BLANK_OPCODE + Chr( nFormat ) + cell_spot, aWks )  // Write blank op
         ENDCASE
         retval := ALL_OK                                  // and cell location

      ELSE
         retval := ERR_INVALID_CELL

      ENDIF
   ENDIF
   RETURN retval
*****************************************************************************
* End of Lblank *
*****************************************************************************


* Function:    Lcreate()
* Purpose:     Creates a new spreadsheet
* Syntax:      Lcreate( cFilename, cVersion )
* Arguments:   cFilename      - Name of new spreadsheet to create
*              cVersion       - Spreadsheet version
*
*              Valid spreadsheet versions are:
*
*              QP  - Quattro Pro     L1  - Lotus 1.x
*              QU  - Quattro         L2  - Lotus 2.x
*                                    L3  - Lotus 3.x
*                                    L4  - Lotus 4.x
*              E2  - Excel 2.x
*              E3  - Excel 3.x
*              E4  - Excel 4.x
*
* Notes:       The Summer '87 version of Lcreate() requires a
*              third paramter, which is a 33 element array.  This
*              array must be declared prior to calling Lcreate().
*              In the Summer '87 version, Lcreate() returns a
*              logical value indicating whether or not the
*              spreadsheet was created.  In Clipper 5, an array
*              is returned.
*
* Returns:     aSpreadsheet
************************************
FUNCTION lcreate( cFilename, cVersion )

   LOCAL nWhere := At( ".", cFilename ), cExt, nHandle, arr_ := { }
   LOCAL nTop    := 3
   LOCAL nLeft   := 3
   LOCAL nPage   := 66
   LOCAL nRight  := 78
   LOCAL nBottom := 3
   DEFAULT cVersion TO "L2"
   IF SubStr( cVersion, 1, 1 ) == "E"
      nTop    := 1
      nLeft   := 1
      nPage   := 66
      nRight  := 1
      nBottom := 1
   ENDIF
   cExt      := Lext( cVersion )
   cFilename := IF( nWhere = 0, Trim( cFilename ) + cExt, cFilename )
   nHandle   := FCreate( cFilename )

   IF nHandle > 0


      arr_ := {  cFilename, ;                               // Spreadsheet name
                cVersion, ;                                // Version code
                nHandle, ;                                 // Handle file opened on
                .T., ;                                     // Updates allowed?
                "NEW", ;                                   // New / old file
                { 1, 1 }, ;                                 // Rows/columns
                { "L", "A", "N", ;
                  IF( SubStr( cVersion, 1, 1 ) == "E", 100, 1 ), ;
                  "P", 114 }, ;                             // Miscellaneous parameters
                .F., ;                                      // Data written flag
                "", ;                                       // Data buffer
                { "", "", "", "", "F", nLeft, nRight, nPage, nTop, nBottom, { "", "" } }, ;
                { "", { "", 0 }, { "", 0 } }, ;                     // Sort array
                { }, ;                                       // Named ranges
                { }, ;                                       // Column widths
                "A1" }                                     // First cell
      IF cVersion $ "L3|L4" .OR. SubStr( cVersion, 1, 1 ) == "E"  // Formats, L3 and
         AAdd( arr_, { } )                                    // Excel, L3 only
      ENDIF
   ENDIF
   RETURN arr_
*****************************************************************************
* End of Lcreate *
*****************************************************************************


*
* Function:    Lfunc()
* Purpose:     Write an @function into the spreadsheet
* Syntax:      Lfunc( aSpreadsheet, cFormula, cCell )
* Arguments:   aSpreadsheet - spreadsheet handle array
*              cFormula     - Formula to write in cell
*              cCell        - Cell address
*
* Notes:       Lfunc() is used mainly for compatibility with
*              earlier CLIPWKS releases.  Lput() handles formula
*              much more powerfully than Lfunc() can.  If you are
*              writing new formulas to the spreadsheet, use
*              Lput() instead.
* Returns:     nStatus
* See Also:    Lput()
**************************************************
FUNCTION lfunc( aWks, cFormula, cCell, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 )

   LOCAL parms := "", arr_ := { arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 }, k

   FOR k = 1 TO Len( arr_ )

      IF ValType( arr_[ k ] ) == "C"
         parms += arr_[ k ] + ","
      ELSEIF ValType( arr_[ k ] ) == "N"
         parms += Str( arr_[ k ] ) + ","
      ENDIF

   NEXT
   parms := IF( SubStr( parms, Len( parms ), 1 ) == ",", SubStr( parms, 1, Len( parms ) - 1 ), parms )
// Build a valid lput() call
   RETURN lput( aWks, "@" + cFormula + "(" + parms + ")", cCell )
******************************************************************************
* End of Lfunc *
******************************************************************************


* Function:    Lput()
* Purpose:     Writes data to the spreadsheet
* Syntax:      Lput( aSpreadsheet, xData, cCell_address ,cFormat )
* Arguments:   aSpreadsheet   - spreadsheet handle array
*              xData          - Data to write
*              cCell_address  - Cell address to write data at
*              cFormat        - Format byte for the cell
* Notes:       The type of data determines what gets written in
*              the cell.  Numeric data is written out directly,
*              logical data is converted to either the @TRUE or
*              @FALSE function, and dates are converted to the
*              @DATE function.  Character data is intrepretted by
*              the first byte.  If the first byte is a plus sign
*              (+) or an ampersand (@), then it is assume to be a
*              formula and the formula will be written out.  Any
*              other first character will result in the data
*              being written directly to the spreadsheet.
*
*              For example:
*
*              Lput(aWks,"+A5*@SUM(B2..B6)","A1")
*              Lput(aWks,date(),"A2")
*              Lput(aWks,PAYROLL->salary,"A3")
*
* Returns:     nStatus
* See Also:    Lblank() Lfunc()
**************************************************
FUNCTION lput( aWks, _data, _cell, cFormat, lWrite )

   LOCAL retval := 0, _string := "", temp := "", places, cell_spot := Lcell( _cell, WKS_VERSION )
   LOCAL lalign := Chr( 94 ), lversion, temp_out := "", _row, _col, p_ := { }, k, r1, c1, nFormat
   LOCAL aQpro := { "", "", "" }, arr_, x, y, nXf := 0
   LOCAL lProtect
   //LOCAL f_h,


   IF ValType( _data ) == "D" .AND. ( WKS_VERSION == "E3" .OR. WKS_VERSION == "E4" )
      nXf := 19
   ELSEIF ValType( _data ) == "D" .AND. WKS_VERSION == "E2"
      nXf := 13
   ENDIF

   IF lWrite == NIL
      lWrite := .T.
   ENDIF

   IF cFormat <> NIL
      IF SubStr( cFormat, 1, 1 ) $ "+-"        // If protection was specified
         lProtect := SubStr( cFormat, 1, 1 ) == "+"
         cFormat  := SubStr( cFormat, 2 )
      ENDIF
      IF ! Empty( cFormat )
         IF SubStr( WKS_VERSION, 1, 1 ) == "E"
            nXf := Lcformat( cFormat, aWks, lProtect )
         ELSE
            nFormat := lcformat( cFormat, aWks, lProtect )
         ENDIF
      ELSE
         nFormat := DEFAULT_FORMAT
         IF nFormat > 127 .AND. ! lProtect
            nFormat -= 128
         ELSEIF nFormat < 128 .AND. lProtect
            nFormat += 128
         ENDIF
      ENDIF
   ELSE
      nFormat := DEFAULT_FORMAT
   ENDIF
   IF lWrite
      retval   := IF( WKS_TYPE == "NEW", ALL_OK, ERR_INVALID_PARAMETERS )
   ENDIF

   IF retval == ALL_OK

      // Since this is only appropiate for character cells, move it to there

      IF lWrite
         lalign    := IF( WKS_ALIGNMENT == 'R', Chr( 34 ), IF( WKS_ALIGNMENT == 'C', Chr( 94 ), Chr( 39 ) ) )
      ENDIF

      // Probably can convert all lversion to WKS_VERSION without any ills

      lversion  := WKS_VERSION
      _row      := Bin2I( SubStr( cell_spot, 1, 2 ) )
      IF lversion $ "L3|L4"
         _col      := Asc( SubStr( cell_spot, 4, 1 ) )
         IF Asc( SubStr( cell_spot, 3, 1 ) ) > nHighest_sheet
            nHighest_sheet := Asc( SubStr( cell_spot, 3, 1 ) )
         ENDIF
      ELSE
         _col      := Bin2I( SubStr( cell_spot, 3, 2 ) )
      ENDIF

      DO CASE
      CASE ValType( _data ) == "C"                   // for characters
         IF SubStr( _data, 1, 1 ) == "@"
            _data := "+" + _data
         ENDIF
         IF SubStr( _data, 1, 1 ) $ "-+@="           // If first character is a +
            IF SubStr( _data, 1, 1 ) $ "+-"
               p_   := Parse( _data )             // or @, assume a formula
               IF ! Empty( p_ )
                  IF WKS_VERSION <> "QP" .AND. SubStr( WKS_VERSION, 1, 1 ) <> "E"
                     FOR k = Len( p_ ) TO 1 STEP - 1
                        IF SubStr( p_[ k ], 1, 1 ) == "@"
                           temp_out += makeformula( p_[ k ], lversion, _row, _col )
                        ELSEIF Val( p_[ k ] ) == 0
                           y := Len( p_[ k ] )
                           IF p_[ k ] <> Replicate( "0", y )
                              temp_out += to_123( p_[ k ], lversion, _row, _col )
                           ELSE
                              temp_out += to_123( Val( p_[ k ] ), lversion, _row, _col )
                           ENDIF
                        ELSE
                           temp_out += to_123( Val( p_[ k ] ), lversion, _row, _col )
                        ENDIF
                     NEXT
                     temp_out += to_123( "ENDMARK", lversion )
                  ELSEIF SubStr( WKS_VERSION, 1, 1 ) == "E"
                     FOR k = Len( p_ ) TO 1 STEP - 1
                        IF SubStr( p_[ k ], 1, 1 ) == "="
                           temp_out += makeformula( p_[ k ], lversion, _row, _col )
                        ELSEIF Val( p_[ k ] ) == 0
                           y := Len( p_[ k ] )
                           IF p_[ k ] <> Replicate( "0", y )
                              temp_out += to_123( p_[ k ], lversion, _row, _col )
                           ELSE
                              temp_out += to_123( Val( p_[ k ] ), lversion, _row, _col )
                           ENDIF
                        ELSE
                           temp_out += to_123( Val( p_[ k ] ), lversion, _row, _col )
                        ENDIF
                     NEXT
                  ELSE
                     // Quattro Pro formula
                     FOR k = Len( p_ ) TO 1 STEP - 1
                        IF SubStr( p_[ k ], 1, 1 ) == "@"
                           arr_ := makeformula( p_[ k ], lversion, _row, _col )
                           aQpro[ QP_FORMULAS ] += arr_[ QP_FORMULAS ]
                           aQpro[ QP_RANGES ]   += arr_[ QP_RANGES ]
                           aQpro[ QP_CELLS ]    += arr_[ QP_CELLS ]
                        ELSEIF Val( p_[ k ] ) == 0
                           IF ".." $ p_[ k ]
                              aQpro[ QP_FORMULAS ] += To_123( "RANGEMARK", "QP", _row, _col )
                              aQpro[ QP_RANGES ] += to_123( p_[ k ], lversion, _row, _col )
                           ELSEIF Len( p_[ k ] ) > 1
                              aQpro[ QP_FORMULAS ] += To_123( "CELL_ADDR", "QP", _row, _col )
                              aQpro[ QP_CELLS ] += to_123( p_[ k ], lversion, _row, _col )
                           ELSE
                              aQpro[ QP_FORMULAS ] += to_123( p_[ k ], lversion, _row, _col )
                           ENDIF
                        ELSE
                           aQpro[ QP_FORMULAS ] += to_123( Val( p_[ k ] ), lversion, _row, _col )
                        ENDIF
                     NEXT
                     aQpro[ QP_FORMULAS ] += to_123( "ENDMARK", lversion )
                     temp_out += I2Bin( Len( aQpro[ QP_FORMULAS ] ) + 4 ) + ;
                                 I2Bin( Len( aQpro[ QP_FORMULAS ] ) + 4 + Len( aQpro[ QP_CELLS ] ) ) + ;
                                 aQpro[ QP_FORMULAS ] + aQpro[ QP_CELLS ] + aQpro[ QP_RANGES ]
                  ENDIF
               ENDIF
            ELSE
               IF WKS_VERSION $ "L3|L4"
                  r1       := Bin2W( SubStr( cell_spot, 1, 2 ) )
                  c1       := Asc( SubStr( cell_spot, 4, 1 ) )
               ELSE
                  r1       := Bin2W( SubStr( cell_spot, 1, 2 ) )
                  c1       := Bin2W( SubStr( cell_spot, 3, 2 ) )
               ENDIF
               IF WKS_VERSION <> "QP"
                  IF SubStr( WKS_VERSION, 1, 1 ) <> "E"
                     temp_out := Makeformula( _data, WKS_VERSION, r1, c1 )
                     temp_out += to_123( "ENDMARK", lversion )
                  ELSE
                     temp_out := Makeformula( _data, WKS_VERSION, r1, c1 )
                  ENDIF
               ELSE
                  // Quattro Pro formula
                  aQpro := Makeformula( _data, WKS_VERSION, r1, c1 )
                  aQpro[ QP_FORMULAS ] += to_123( "ENDMARK", lversion )
                  temp_out += I2Bin( Len( aQpro[ QP_FORMULAS ] ) + 4 ) + ;
                              I2Bin( Len( aQpro[ QP_FORMULAS ] ) + 4 + Len( aQpro[ QP_CELLS ] ) ) + ;
                              aQpro[ QP_FORMULAS ] + aQpro[ QP_CELLS ] + aQpro[ QP_RANGES ]
               ENDIF
            ENDIF
            IF ! Empty( temp_out )
               IF WKS_VERSION $ "L3|L4"
                  _string := WK3_FORMULACELL + I2Bin( 14 + Len( temp_out ) ) + ;
                             cell_spot + treal( 0 ) + temp_out
               ELSEIF WKS_VERSION == "E2"
                  _string := E2_FORMULA + I2Bin( 17 + Len( temp_out ) ) + ;
                             cell_spot + Chr( 0 ) + Chr( nXf ) + Chr( 0 ) + ieereal( 0 ) + Chr( 1 ) + Chr( Len( temp_out ) ) + ;
                             temp_out
               ELSEIF WKS_VERSION == "E3"
                  _string := E3_FORMULA + I2Bin( 18 + Len( temp_out ) ) + ;
                             cell_spot + I2Bin( nXf ) + ieereal( 0 ) + I2Bin( 1 ) + I2Bin( Len( temp_out ) ) + ;
                             temp_out
               ELSEIF WKS_VERSION == "E4"
                  _string := E4_FORMULA + I2Bin( 18 + Len( temp_out ) ) + ;
                             cell_spot + I2Bin( nXf ) + ieereal( 0 ) + I2Bin( 1 ) + I2Bin( Len( temp_out ) ) + ;
                             temp_out
               ELSE
                  _string := FORMULA_OPCODE + I2Bin( 15 + Len( temp_out ) ) + Chr( nFormat ) + ;
                             cell_spot + ieereal( 0 ) + I2Bin( Len( temp_out ) ) + temp_out
               ENDIF
            ENDIF

         ELSEIF SubStr( _data, 1, 1 ) $ Chr( 34 ) + "\'^|"

            IF WKS_VERSION == "QP"

               _data  := IF( Len( _data ) <= 240, _data, SubStr( _data, 1, 240 ) )
               _string := LABEL_OPCODE + I2Bin( Len( _data ) + 7 ) + Chr( nFormat ) + cell_spot + ;
                          SubStr( _data, 1, 1 ) + Chr( Len( _data ) - 1 ) + SubStr( _data, 2 ) + Chr( 1 )

            ELSEIF WKS_VERSION $ "L3|L4"

               _data  := IF( Len( _data ) <= 512, _data, SubStr( _data, 1, 512 ) )
               _string := WK3_LABELCELL + I2Bin( Len( _data ) + 5 ) + cell_spot + _data + Chr( 0 )

            ELSEIF WKS_VERSION == "E2"
               _data  := IF( Len( _data ) <= 255, _data, SubStr( _data, 1, 255 ) )
               _string := E2_LABEL + I2Bin( Len( _data ) + 7 ) + cell_spot + Chr( 0 ) + I2Bin( 0 ) + Chr( Len( _data ) - 1 ) + SubStr( _data, 2 )

            ELSEIF WKS_VERSION == "E3" .OR. WKS_VERSION == "E4"
               x := At( SubStr( _data, 1, 1 ), ['^"\] )
               x := 0
               _data  := IF( Len( _data ) <= 255, _data, SubStr( _data, 1, 255 ) )
               _string := E3_LABEL + I2Bin( Len( _data ) + 7 ) + cell_spot + ;
                          I2Bin( x ) + I2Bin( Len( _data ) - 1 ) + SubStr( _data, 2 )

            ELSE
               _data  := IF( Len( _data ) <= 241, _data, SubStr( _data, 1, 241 ) )
               _string := LABEL_OPCODE + I2Bin( Len( _data ) + 6 ) + Chr( nFormat ) + cell_spot + _data + Chr( 0 )
            ENDIF

         ELSE
            IF WKS_VERSION == "QP"

               _data  := IF( Len( _data ) <= 240, _data, SubStr( _data, 1, 240 ) )
               _string := LABEL_OPCODE + I2Bin( Len( _data ) + 8 ) + Chr( nFormat ) + cell_spot + lalign + ;
                          Chr( Len( _data ) ) + _data + Chr( 1 )

            ELSEIF WKS_VERSION $ "L3|L4"

               _data  := IF( Len( _data ) <= 512, _data, SubStr( _data, 1, 512 ) )
               _string := WK3_LABELCELL + I2Bin( Len( _data ) + 6 ) + cell_spot + "'" + _data + Chr( 0 )

            ELSEIF WKS_VERSION == "E2"
               _data  := IF( Len( _data ) <= 255, _data, SubStr( _data, 1, 255 ) )
               _string := E2_LABEL + I2Bin( Len( _data ) + 8 ) + cell_spot + Chr( 0 ) + I2Bin( 0 ) + Chr( Len( _data ) ) + _data
            ELSEIF WKS_VERSION == "E3" .OR. WKS_VERSION == "E4"
               _data  := IF( Len( _data ) <= 255, _data, SubStr( _data, 1, 255 ) )
               _string := E3_LABEL + I2Bin( Len( _data ) + 8 ) + cell_spot + I2Bin( 0 ) + I2Bin( Len( _data ) ) + _data

            ELSE

               _data  := IF( Len( _data ) <= 240, _data, SubStr( _data, 1, 240 ) )
               _string := LABEL_OPCODE + I2Bin( Len( _data ) + 7 ) + Chr( nFormat ) + cell_spot + lalign + _data + Chr( 0 )
            ENDIF
         ENDIF

      CASE ValType( _data ) == "L"        // Logicals ******************************

         IF WKS_VERSION == "E2"
            _string := E2_BOOLERR + cell_spot + Chr( 0 ) + Chr( nXf ) + Chr( 0 ) + ;
                       Chr( IF( _data, 1, 0 ) ) + Chr( 0 )
         ELSEIF WKS_VERSION == "E3"
            _string := E3_FORMULA + I2Bin( 20 ) + cell_spot + I2Bin( 0 ) + ;
                       ieereal( IF( _data, 1, 0 ) ) + I2Bin( 0 ) + ;
                       I2Bin( 2 ) + Chr( 29 ) + Chr( IF( _data, 1, 0 ) )
         ELSEIF WKS_VERSION == "E4"
            _string := E4_FORMULA + I2Bin( 20 ) + cell_spot + I2Bin( 0 ) + ;
                       ieereal( IF( _data, 1, 0 ) ) + I2Bin( 0 ) + ;
                       I2Bin( 2 ) + Chr( 29 ) + Chr( IF( _data, 1, 0 ) )
         ELSE
            IF _data
               RETURN lput( aWks, "@TRUE", _cell )
            ELSE
               RETURN lput( aWks, "@FALSE", _cell )
            ENDIF
         ENDIF

      CASE ValType( _data ) == "D"        // Dates *********************************

         IF WKS_VERSION == "E2"
            _string := E2_NUMBER + cell_spot + Chr( 0 ) + Chr( nXf ) + Chr( 0 ) + ieereal( _data - CToD( "12/31/1899" ) )
         ELSEIF WKS_VERSION == "E3" .OR. WKS_VERSION == "E4"
            _string := E3_NUMBER + cell_spot + I2Bin( nXf ) + ieereal( _data - CToD( "12/30/1899" ) )
         ELSE
            cFormat := IF( cFormat == NIL, "D1", cFormat )
            IF ! Empty( _data )
               RETURN lput( aWks, "@DATE(" + AllTrim( Str( Year( _data ) - 1900, 4 ) ) + "," + ;
                                          Str( Month( _data ), 2 ) + "," + ;
                                          Str( Day( _data ), 2 ) + ")", _cell, cFormat )
            ELSE
               RETURN lput( aWks, "@DATE(-1,-1,-1)", _cell, cFormat )
            ENDIF
         ENDIF

      CASE ValType( _data ) == "N"        // Numeric data
         temp   := Str( _data )
         places := IF( At( ".", temp ) = 0, 0, Len( temp ) - At( ".", temp ) )
         IF WKS_VERSION $ "L3|L4"
            IF _data >= - 16384 .AND. _data <= 16383 .AND. places == 0
               _string := WK3_SMALLNUMCELL + cell_spot + shortnum( _data )
            ELSE
               _string := WK3_NUMBERCELL + cell_spot + treal( _data )
            ENDIF
         ELSEIF WKS_VERSION == "E2"
            _string := E2_NUMBER + cell_spot + Chr( 0 ) + Chr( nXf ) + Chr( 0 ) + ieereal( _data )
         ELSEIF WKS_VERSION == "E3" .OR. WKS_VERSION == "E4"
            _string := E3_NUMBER + cell_spot + I2Bin( nXf ) + ieereal( _data )

         ELSE
            IF nFormat == DEFAULT_FORMAT
               IF places <> 0
                  _string := REAL_OPCODE + Chr( 128 + places ) + cell_spot + ieereal( _data )
               ELSE
                  _string := REAL_OPCODE + Chr( 192 ) + cell_spot + ieereal( _data )
               ENDIF
            ELSE
               _string := REAL_OPCODE + Chr( nFormat ) + cell_spot + ieereal( _data )
            ENDIF
         ENDIF
      ENDCASE
      IF lWrite
         Bwrite( _string, aWks )
      ELSE
         RETURN _string
      ENDIF
   ENDIF
   RETURN retval
*****************************************************************************
* End of Lput *
*****************************************************************************

* Function:    Lwidth()
* Purpose:     Set the column to the specified width
* Syntax:      Lwidth( aSpreadsheet, cColumn|nColumn, nWidth )
* Arguments:   aSpreadsheet - Spreadsheet handle array
*              cColumn      - Column letter
*              nColumn      - Numeric letter
*              nWidth       - Width to set
*              nSheet       - Sheet to set width for
* Notes:       Lwidth() is used to set a column width.  You may
*              specify the column using letter notation or by
*              refering to the column number.  Widths should be
*              set prior to any data being written to the
*              spreadsheet.
* Returns:     nStatus
**************************************************
FUNCTION lwidth( aWks, xColumn, nWidth, nSheet )

   LOCAL retval := IF( ValType( WKS_TYPE ) == "C", ALL_OK, ERR_INVALID_PARAMETERS )
   LOCAL _hold
   LOCAL x

   DEFAULT nWidth  TO  DEFAULT_WIDTH
   DEFAULT nSheet  TO  0

   IF xColumn == NIL
      xColumn := 0
   ENDIF

   IF retval == ALL_OK
      // Convert alphabetic column references to numeric
      IF ValType( xColumn ) == "C"
         _hold := Upper( Trim( xColumn ) )
         IF ":" $ _hold .AND. WKS_VERSION $ "L3|L4"
            x      := At( ":", _hold )
            IF x = 2
               nSheet := Asc( _hold ) - 65
               _hold  := SubStr( _hold, 3 )
            ELSE
               nSheet := ( Asc( _hold ) - 64 ) * 26 + ( Asc( SubStr( _hold, 2, 1 ) ) - 65 )
               _hold  := SubStr( _hold, 4 )
            ENDIF
         ENDIF
         IF Len( _hold ) = 1 .AND. _hold $ "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            xColumn := Asc( _hold ) - 64
         ELSEIF Len( _hold ) = 2 .AND. ;
            SubStr( _hold, 1, 1 ) $ "ABCDEFGHI" .AND. ;
            SubStr( _hold, 2, 1 ) $ "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            xColumn := ( Asc( SubStr( _hold, 1, 1 ) ) - 64 ) * 26 + ;
                       ( Asc( SubStr( _hold, 2, 1 ) ) - 64 )
         ENDIF
      ENDIF
      IF ValType( xColumn ) = "N"
         IF xColumn = 0
            IF ! ( WKS_VERSION $ "L3|L4" )
               AAdd( WKS_WIDTHS, { 9999, nWidth } )
            ENDIF
         ELSE
            IF WKS_VERSION $ "L3|L4"
               IF nSheet > nHighest_sheet
                  nHighest_sheet := nSheet
               ENDIF
               AAdd( WKS_WIDTHS, Str( nSheet - 1, 3 ) + Chr( xColumn - 1 ) + Chr( nWidth ) )
            ELSE
               AAdd( WKS_WIDTHS, { xColumn - 1, nWidth } )
            ENDIF
         ENDIF
         retval := ALL_OK
      ENDIF
   ENDIF
   RETURN retval
*****************************************************************************
* End of Lwidth *
*****************************************************************************


* Function:    Ldirty()
* Purpose:     Writes the global parameters to disk before data is written
* Syntax:      Ldirty( aSpreadsheet )
* Arguments:   aSpreadsheet - Spreadsheet handle array
* Notes:       CLIPWKS maintains all global parameters in the
*              spreadsheet array until you request an Lput()
*              function.  Once you call Lput(), the global
*              information is formatted and written to the
*              spreadsheet file.  You should never need to call
*              Ldirty() directly.
* Returns:     NIL
* Class:       Internal
**************************************************
FUNCTION ldirty( aWks )

   LOCAL buffer   := "", jj, nCol, nWide, cName, cRange, tmp, outs, x, ct, nRow, nSheet, zz
   LOCAL cExAlign := ""

   IF ! WKS_DATA_WRITTEN                        // Parameter data not yet saved
******************************************************************************
* Place a BOF op code into the file *
******************************************************************************
      DO CASE
      CASE WKS_VERSION == "L1"
         buffer := BOF_OPCODE + I2Bin( 1028 )
      CASE WKS_VERSION == "L2"
         buffer := BOF_OPCODE + I2Bin( 1030 )
      CASE WKS_VERSION == "L3"
         buffer = WK3_BOF + I2Bin( 0 ) + I2Bin( 0 ) + I2Bin( WKS_ROWS - 1 ) + ;
         Chr( nHighest_sheet - 1 ) + Chr( WKS_COLUMNS - 1 ) + I2Bin( 1 ) + I2Bin( 0 ) + ;
         Chr( 1 ) + Chr( 2 ) + Replicate( Chr( 0 ), 8 )
      CASE WKS_VERSION == "L4"
         buffer = WK4_BOF + I2Bin( 0 ) + I2Bin( 0 ) + I2Bin( WKS_ROWS - 1 ) + ;
         Chr( nHighest_sheet ) + Chr( WKS_COLUMNS - 1 ) + I2Bin( 1 ) + I2Bin( 0 ) + ;
         Chr( 1 ) + Chr( 2 ) + Replicate( Chr( 0 ), 8 )
      CASE WKS_VERSION == "QU"
         buffer := BOF_OPCODE + I2Bin( 2571 )
      CASE WKS_VERSION == "QP"
         buffer := BOF_OPCODE + I2Bin( 20768 )
      CASE WKS_VERSION == "E2"
         buffer := E2_BOF_OPCODE
      CASE WKS_VERSION == "E3"
         buffer := E3_BOF_OPCODE
      CASE WKS_VERSION == "E4"
         buffer := E4_BOF_OPCODE
      ENDCASE
******************************************************************************
* Excel 2.x *
******************************************************************************
      IF WKS_VERSION == "E2"
         buffer += E2_CODEPAGE        + I2Bin( 32769 )
         buffer += E2_CALCCOUNT       + I2Bin( WKS_ITERATION )
         buffer += E2_CALCMODE        + I2Bin( IF( WKS_CALCMODE == "M", 0, IF( WKS_CALCMODE == "T", - 1, 1 ) ) )

         buffer += E2_PRECISION       + I2Bin( 1 )
         buffer += E2_REFMODE         + I2Bin( 1 )
         buffer += E2_DELTA           + "MbP?"
         buffer += E2_ITERATION       + I2Bin( 0 )
         buffer += E2_1904            + I2Bin( 0 )
         buffer += E2_PRINT_HEADERS   + I2Bin( IF( WKS_PRINT == "F", 1, 0 ) )
         buffer += E2_PRINT_GRIDLINES + I2Bin( IF( WKS_PRINT == "F", 1, 0 ) )
         buffer += E2_PROTECT         + I2Bin( IF( WKS_PROTECTION == "P", 1, 0 ) )
         IF Empty( WKS_HEADER )
            buffer += E2_HEADER + I2Bin( 3 ) + Chr( 2 ) + "&f"
         ELSE
            buffer += E2_HEADER + I2Bin( 1 + Len( WKS_HEADER ) ) + Chr( Len( WKS_HEADER ) ) + WKS_HEADER
         ENDIF
         IF Empty( WKS_FOOTER )
            buffer += E2_FOOTER + I2Bin( 8 ) + Chr( 7 ) + "Page &p"
         ELSE
            buffer += E2_FOOTER + I2Bin( 1 + Len( WKS_FOOTER ) ) + Chr( Len( WKS_FOOTER ) ) + WKS_FOOTER
         ENDIF
         buffer += E2_LEFT_MARGIN     + ieereal( WKS_LEFT )
         buffer += E2_RIGHT_MARGIN    + ieereal( WKS_RIGHT )
         buffer += E2_TOP_MARGIN      + ieereal( WKS_TOP )
         buffer += E2_BOTTOM_MARGIN   + ieereal( WKS_BOTTOM )
         buffer += E2_BACKUP          + I2Bin( 0 )
         buffer += E3_FORMAT + I2Bin( 8 ) + Chr( 7 ) + "General"
         buffer += E3_FORMAT + I2Bin( 2 ) + Chr( 1 ) + "0"
         buffer += E3_FORMAT + I2Bin( 5 ) + Chr( 4 ) + "0.00"
         buffer += E3_FORMAT + I2Bin( 6 ) + Chr( 5 ) + "#,##0"
         buffer += E3_FORMAT + I2Bin( 9 ) + Chr( 8 ) + "#,##0.00"
         buffer += E3_FORMAT + I2Bin( 22 ) + Chr( 21 ) + '"$"#,##0;\("$"#,##0\)'
         buffer += E3_FORMAT + I2Bin( 27 ) + Chr( 26 ) + '"$"#,##0;[Red]\("$"#,##0\)'
         buffer += E3_FORMAT + I2Bin( 28 ) + Chr( 27 ) + '"$"#,##0.00;\("$"#,##0.00\)'
         buffer += E3_FORMAT + I2Bin( 33 ) + Chr( 32 ) + '"$"#,##0.00;[Red]\("$"#,##0.00\)'
         buffer += E3_FORMAT + I2Bin( 3 ) + Chr( 2 ) + "0%"
         buffer += E3_FORMAT + I2Bin( 6 ) + Chr( 5 ) + "0.00%"
         buffer += E3_FORMAT + I2Bin( 9 ) + Chr( 8 ) + "0.00E+00"
         buffer += E3_FORMAT + I2Bin( 7 ) + Chr( 6 ) + "m/d/yy"
         buffer += E3_FORMAT + I2Bin( 11 ) + Chr( 10 ) + "d\-mmm\-yy"
         buffer += E3_FORMAT + I2Bin( 7 ) + Chr( 6 ) + "d\-mmm"
         buffer += E3_FORMAT + I2Bin( 8 ) + Chr( 7 ) + "mmm\-yy"
         buffer += E3_FORMAT + I2Bin( 12 ) + Chr( 11 ) + "h:mm\ AM/PM"
         buffer += E3_FORMAT + I2Bin( 15 ) + Chr( 14 ) + "h:mm:ss\ AM/PM"
         buffer += E3_FORMAT + I2Bin( 5 ) + Chr( 4 ) + "h:mm"
         buffer += E3_FORMAT + I2Bin( 8 ) + Chr( 7 ) + "h:mm:ss"
         buffer += E3_FORMAT + I2Bin( 13 ) + Chr( 12 ) + "m/d/yy\ h:mm"
         buffer += E2_DEFAULTWIDTH + I2Bin( 8 )
         FOR x = 1 TO Len( WKS_WIDTHS )
            IF WKS_WIDTHS[ x, 1 ] < 9999
               buffer += E2_COLWIDTH + Chr( WKS_WIDTHS[ x, 1 ] ) + Chr( WKS_WIDTHS[ x, 1 ] ) + ;
                         I2Bin( WKS_WIDTHS[ x, 2 ] * 256 )
            ENDIF
         NEXT
         buffer += E2_DIMENSIONS      + I2Bin( 0 ) + I2Bin( WKS_ROWS + 1 ) + ;
                   I2Bin( 0 ) + I2Bin( WKS_COLUMNS + 1 )


******************************************************************************
* Excel 3.x and 4.x *
******************************************************************************
      ELSEIF WKS_VERSION == "E3" .OR. WKS_VERSION == "E4"
         buffer += E3_WRITEACCESS + Chr( 15 ) + PadR( "Markus Wiederstein Excel Format", 31 )
         IF WKS_VERSION == "E3"
            buffer += E2_CODEPAGE  + I2Bin( 32769 )
         ELSE
            buffer += E2_CODEPAGE  + I2Bin( 1252 )
         ENDIF
         buffer += E2_CALCCOUNT + I2Bin( WKS_ITERATION )
         buffer += E2_CALCMODE + I2Bin( IF( WKS_CALCMODE == "M", 0, IF( WKS_CALCMODE == "T", - 1, 1 ) ) )
         buffer += E2_PRECISION + I2Bin( 1 )
         buffer += E2_REFMODE + I2Bin( 1 )
         buffer += E2_DELTA + "MbP?"
         buffer += E2_ITERATION + I2Bin( 0 )
         buffer += E2_1904 + I2Bin( 0 )
         buffer += E2_PRINT_HEADERS + I2Bin( IF( WKS_PRINT == "F", 1, 0 ) )
         buffer += E2_PRINT_GRIDLINES + I2Bin( IF( WKS_PRINT == "F", 1, 0 ) )
         buffer += E3_SAVERECALC + I2Bin( 0 )
         buffer += E3_GRIDSET + I2Bin( IF( WKS_PRINT == "F", 1, 0 ) )
         buffer += E3_HCENTER + I2Bin( IF( WKS_PRINT == "F", 1, 0 ) )
         buffer += E3_VCENTER + I2Bin( IF( WKS_PRINT == "F", 1, 0 ) )
         buffer += E3_GUTS + I2Bin( 0 ) + I2Bin( 0 )
         buffer += E3_DEFAULT_HEIGHT + I2Bin( 0 ) + I2Bin( 255 )
         buffer += E3_COUNTRY + I2Bin( 1 ) + I2Bin( 1 )
         buffer += E3_HIDEOBJ + I2Bin( 0 )
         buffer += E3_WSBOOL + Chr( 193 ) + Chr( 4 )
         FOR x = 0 TO 3
            buffer += E3_FONT + I2Bin( 20 ) + I2Bin( 200 ) + I2Bin( x ) + Chr( 255 ) + Chr( 127 ) + Chr( 13 ) + "MS Sans Serif"
         NEXT
         buffer += E3_FONT + I2Bin( 20 ) + I2Bin( 200 ) + I2Bin( 0 ) + Chr( 255 ) + Chr( 127 ) + Chr( 13 ) + "MS Sans Serif"

         IF Empty( WKS_HEADER )
            buffer += E2_HEADER + I2Bin( 3 ) + Chr( 2 ) + "&f"
         ELSE
            buffer += E2_HEADER + I2Bin( 1 + Len( WKS_HEADER ) ) + Chr( Len( WKS_HEADER ) ) + WKS_HEADER
         ENDIF
         IF Empty( WKS_FOOTER )
            buffer += E2_FOOTER + I2Bin( 8 ) + Chr( 7 ) + "Page &p"
         ELSE
            buffer += E2_FOOTER + I2Bin( 1 + Len( WKS_FOOTER ) ) + Chr( Len( WKS_FOOTER ) ) + WKS_FOOTER
         ENDIF
         buffer += E2_LEFT_MARGIN + ieereal( WKS_LEFT )
         buffer += E2_RIGHT_MARGIN + ieereal( WKS_RIGHT )
         buffer += E2_TOP_MARGIN + ieereal( WKS_TOP )
         buffer += E2_BOTTOM_MARGIN + ieereal( WKS_BOTTOM )
         buffer += E2_BACKUP + I2Bin( 0 )
         buffer += E2_BUILTINFORMAT + I2Bin( 23 )



         buffer += E3_FORMAT + I2Bin( 8 ) + Chr( 7 ) + "General"
         buffer += E3_FORMAT + I2Bin( 2 ) + Chr( 1 ) + "0"
         buffer += E3_FORMAT + I2Bin( 5 ) + Chr( 4 ) + "0.00"
         buffer += E3_FORMAT + I2Bin( 6 ) + Chr( 5 ) + "#,##0"
         buffer += E3_FORMAT + I2Bin( 9 ) + Chr( 8 ) + "#,##0.00"
         buffer += E3_FORMAT + I2Bin( 24 ) + Chr( 23 ) + '"$"#,##0_);\("$"#,##0\)'
         buffer += E3_FORMAT + I2Bin( 29 ) + Chr( 28 ) + '"$"#,##0_);[Red]\("$"#,##0\)'
         buffer += E3_FORMAT + I2Bin( 30 ) + Chr( 29 ) + '"$"#,##0.00_);\("$"#,##0.00\)'
         buffer += E3_FORMAT + I2Bin( 35 ) + Chr( 34 ) + '"$"#,##0.00_);[Red]\("$"#,##0.00\)'
         buffer += E3_FORMAT + I2Bin( 3 ) + Chr( 2 ) + "0%"
         buffer += E3_FORMAT + I2Bin( 6 ) + Chr( 5 ) + "0.00%"
         buffer += E3_FORMAT + I2Bin( 9 ) + Chr( 8 ) + "0.00E+00"
         buffer += E3_FORMAT + I2Bin( 7 ) + Chr( 6 ) + "#\ ?/?"
         buffer += E3_FORMAT + I2Bin( 9 ) + Chr( 8 ) + "#\ ??/??"
         buffer += E3_FORMAT + I2Bin( 7 ) + Chr( 6 ) + "m/d/yy"
         buffer += E3_FORMAT + I2Bin( 11 ) + Chr( 10 ) + "d\-mmm\-yy"
         buffer += E3_FORMAT + I2Bin( 7 ) + Chr( 6 ) + "d\-mmm"
         buffer += E3_FORMAT + I2Bin( 8 ) + Chr( 7 ) + "mmm\-yy"
         buffer += E3_FORMAT + I2Bin( 12 ) + Chr( 11 ) + "h:mm\ AM/PM"
         buffer += E3_FORMAT + I2Bin( 15 ) + Chr( 14 ) + "h:mm:ss\ AM/PM"
         buffer += E3_FORMAT + I2Bin( 5 ) + Chr( 4 ) + "h:mm"
         buffer += E3_FORMAT + I2Bin( 8 ) + Chr( 7 ) + "h:mm:ss"
         buffer += E3_FORMAT + I2Bin( 13 ) + Chr( 12 ) + "m/d/yy\ h:mm"

         buffer += E3_NAME + I2Bin( 7 ) + I2Bin( 32 ) + Chr( 47 ) + Chr( 1 ) + I2Bin( 0 ) + Chr( 5 )
         buffer += E3_NAME + I2Bin( 7 ) + I2Bin( 32 ) + Chr( 47 ) + Chr( 1 ) + I2Bin( 0 ) + Chr( 4 )
         buffer += E3_NAME + I2Bin( 7 ) + I2Bin( 32 ) + Chr( 47 ) + Chr( 1 ) + I2Bin( 0 ) + Chr( 9 )
         buffer += E3_NAME + I2Bin( 7 ) + I2Bin( 32 ) + Chr( 47 ) + Chr( 1 ) + I2Bin( 0 ) + Chr( 6 )
         buffer += E3_NAME + I2Bin( 7 ) + I2Bin( 32 ) + Chr( 47 ) + Chr( 1 ) + I2Bin( 0 ) + Chr( 7 )
         buffer += E3_NAME + I2Bin( 7 ) + I2Bin( 32 ) + Chr( 47 ) + Chr( 1 ) + I2Bin( 0 ) + Chr( 8 )

         FOR x := 1 TO Len( EXCEL_NAMED_RANGES )
            zz := Len( AllTrim( EXCEL_NAMED_RANGES[ x, 1 ] ) )
            buffer += E3_NAME + I2Bin( 13 + zz ) + I2Bin( 0 ) + "/" + ;
                      Chr( zz ) + I2Bin( 7 ) + EXCEL_NAMED_RANGES[ x, 1 ] + Chr( 45 ) + ;
                      Left( EXCEL_NAMED_RANGES[ x, 2 ], 2 ) + ;
                      SubStr( EXCEL_NAMED_RANGES[ x, 2 ], 5, 2 ) + ;
                      SubStr( EXCEL_NAMED_RANGES[ x, 2 ], 3, 1 ) + ;
                      SubStr( EXCEL_NAMED_RANGES[ x, 2 ], 7, 1 )
         NEXT

         buffer += E2_PROTECT + I2Bin( IF( WKS_PROTECTION == "P", 1, 0 ) )
         buffer += I2Bin( 25 ) + I2Bin( 2 ) + I2Bin( 0 )
         buffer += I2Bin( 99 ) + I2Bin( 2 ) + I2Bin( 0 )
         buffer += I2Bin( 19 ) + I2Bin( 2 ) + I2Bin( 0 )

         IF WKS_ALIGNMENT = "R"
            cExAlign := Chr( 243 )
         ENDIF
         IF WKS_ALIGNMENT = "C"
            cExAlign := Chr( 242 )
         ENDIF

         buffer += XF_OPCODE + I2Bin( 0 ) + I2Bin( 5 ) + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + I2Bin( 1 ) + Chr( 5 ) + "" + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + I2Bin( 1 ) + Chr( 5 ) + "" + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + I2Bin( 2 ) + Chr( 5 ) + "" + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + I2Bin( 2 ) + Chr( 5 ) + "" + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         FOR x = 1 TO 10
            buffer += XF_OPCODE + I2Bin( 0 ) + Chr( 5 ) + "" + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         NEXT
         buffer += XF_OPCODE + I2Bin( 0 ) + I2Bin( 1 ) + I2Bin( 0 ) + Chr( 0 ) + Chr( 206 ) + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 5 ) + Chr( 4 ) + Chr( 5 ) + "" + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 5 ) + Chr( 8 ) + Chr( 5 ) + "" + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 5 ) + Chr( 9 ) + Chr( 5 ) + "" + cExAlign + "" + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         //
         // Special XF records for D1,D2,D3,Percent,comma,currency,sci,fix
         //
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 14 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 17 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 16 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 9 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 10 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 3 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 4 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 5 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 7 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 11 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 1 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )
         buffer += XF_OPCODE + Chr( 0 ) + Chr( 2 ) + Chr( 1 ) + Chr( 4 ) + I2Bin( 0 ) + Chr( 0 ) + "" + I2Bin( 0 ) + I2Bin( 0 )

         buffer += E3_STYLE + Chr( 16 ) + Chr( 128 ) + Chr( 3 ) + Chr( 255 )
         buffer += E3_STYLE + Chr( 17 ) + Chr( 128 ) + Chr( 4 ) + Chr( 255 )
         buffer += E3_STYLE + Chr( 0 ) + Chr( 128 ) + Chr( 0 ) + Chr( 255 )
         buffer += E3_STYLE + Chr( 18 ) + Chr( 128 ) + Chr( 5 ) + Chr( 255 )
         buffer += E2_DEFAULTWIDTH + I2Bin( 8 )
         FOR x = 1 TO Len( WKS_WIDTHS )
            IF WKS_WIDTHS[ x, 1 ] < 9999
               buffer += E3_COLINFO + I2Bin( WKS_WIDTHS[ x, 1 ] ) + I2Bin( WKS_WIDTHS[ x, 1 ] ) + ;
                         I2Bin( WKS_WIDTHS[ x, 2 ] * 256 ) + I2Bin( 0 ) + I2Bin( 0 ) + Chr( 0 )
            ENDIF
         NEXT


         buffer += E3_DIMENSIONS + I2Bin( 0 ) + I2Bin( WKS_ROWS ) + I2Bin( 0 ) + I2Bin( WKS_COLUMNS ) + I2Bin( 0 )

******************************************************************************
* Other spreadsheets *
******************************************************************************
      ELSEIF ! ( WKS_VERSION $ "L3|L4" )
         // RANGE op code
         buffer += RANGE_OPCODE + I2Bin( WKS_ROWS - 1 ) + I2Bin( WKS_COLUMNS - 1 )

         // Iteration
         buffer += ITERATE_OPCODE + Chr( WKS_ITERATION )

         // Calc mode
         buffer += MODE_OPCODE + IF( WKS_CALCMODE == "A", Chr( 255 ), ;
                                     IF( WKS_CALCMODE == "B", Chr( 1 ), Chr( 0 ) ) )
         // Calc order
         buffer += ORDER_OPCODE + IF( WKS_CALCORDER == "N", Chr( 0 ), ;
                                      IF( WKS_CALCORDER == "C", Chr( 1 ), Chr( 255 ) ) )

         //
         zz := AScan( WKS_WIDTHS, { | z | z[ 1 ] = 9999 } )
         IF zz > 0
            zz := WKS_WIDTHS[ zz, 2 ]
         ELSE
            zz := 9
         ENDIF
         IF WKS_VERSION == "QP"
            buffer += WINDOW1_OPCODE + I2Bin( 38 ) + Lcell( WKS_FIRST_CELL, "QP" ) + ;
                      Chr( 241 ) + Chr( 0 ) + I2Bin( zz ) + I2Bin( 8 ) + I2Bin( 20 ) + ;
                      Lcell( "A1", "QP" ) + I2Bin( 0 ) + I2Bin( 0 ) + Lcell( "A1", "QP" ) + ;
                      I2Bin( 4 ) + I2Bin( 1 ) + I2Bin( 72 ) + Chr( 0 ) + Chr( 0 ) + I2Bin( 20 ) + ;
                      I2Bin( 0 ) + Chr( 1 ) + Chr( 0 )
         ELSEIF WKS_FIRST_CELL <> "A1"
            buffer += SPLIT_OPCODE + Chr( 0 )
            buffer += SYNC_OPCODE + Chr( 255 )
            IF WKS_VERSION <> "L1"
               buffer += WINDOW1_OPCODE + I2Bin( 32 ) + Lcell( WKS_FIRST_CELL, WKS_VERSION ) + ;
                         Chr( 113 ) + Chr( 0 ) + I2Bin( zz ) + I2Bin( 8 ) + I2Bin( 20 ) + ;
                         Lcell( "A1", "L2" ) + I2Bin( 0 ) + I2Bin( 0 ) + Lcell( "A1", "L2" ) + ;
                         Chr( 4 ) + Chr( 0 ) + Chr( 4 ) + Chr( 0 )
               buffer += HIDVEC1_OPCODE + Replicate( Chr( 0 ), 32 )
            ELSE
               buffer += WINDOW1_OPCODE + I2Bin( 31 ) + Lcell( WKS_FIRST_CELL, WKS_VERSION ) + ;
                         Chr( 113 ) + Chr( 0 ) + I2Bin( zz ) + I2Bin( 8 ) + I2Bin( 20 ) + ;
                         Lcell( "A1", "L2" ) + I2Bin( 0 ) + I2Bin( 0 ) + Lcell( "A1", "L2" ) + ;
                         Chr( 4 ) + Chr( 0 ) + Chr( 0 )
            ENDIF
         ELSE
            IF WKS_VERSION <> "L1"
               buffer += WINDOW1_OPCODE + I2Bin( 32 ) + Lcell( WKS_FIRST_CELL, WKS_VERSION ) + ;
                         Chr( 113 ) + Chr( 0 ) + I2Bin( zz ) + I2Bin( 8 ) + I2Bin( 20 ) + ;
                         Lcell( "A1", "L2" ) + I2Bin( 0 ) + I2Bin( 0 ) + Lcell( "A1", "L2" ) + ;
                         Chr( 4 ) + Chr( 0 ) + Chr( 4 ) + Chr( 0 )
               buffer += HIDVEC1_OPCODE + Replicate( Chr( 0 ), 32 )
            ELSE
               buffer += WINDOW1_OPCODE + I2Bin( 31 ) + Lcell( WKS_FIRST_CELL, WKS_VERSION ) + ;
                         Chr( 113 ) + Chr( 0 ) + I2Bin( zz ) + I2Bin( 8 ) + I2Bin( 20 ) + ;
                         Lcell( "A1", "L2" ) + I2Bin( 0 ) + I2Bin( 0 ) + Lcell( "A1", "L2" ) + ;
                         Chr( 4 ) + Chr( 0 ) + Chr( 0 )
            ENDIF

         ENDIF

         // Column widths
         FOR jj = 1 TO Len( WKS_WIDTHS )
            nCol  := WKS_WIDTHS[ jj ][ 1 ]
            IF nCol <> 9999
               nWide := WKS_WIDTHS[ jj ][ 2 ]
               buffer += WIDTH_OPCODE + I2Bin( nCol ) + Chr( nWide )
            ENDIF
         NEXT
         // Named ranges
         //
         // Need to be sorted, thanks David Kelly for pointing this out
         //
         ASort( WKS_NAMED_RANGES,,, { | x, y | x[ 1 ] < y[ 1 ] } )

         FOR jj = 1 TO Len( WKS_NAMED_RANGES )
            cName   := WKS_NAMED_RANGES[ jj ][ 1 ]
            cRange  := WKS_NAMED_RANGES[ jj ][ 2 ]
            IF WKS_VERSION == "QP"
               buffer  += NAMED_RANGE_OPCODE + Chr( Len( Trim( cName ) ) ) + ;
                          PadR( Trim( cName ), 15 ) + cRange
            ELSE
               buffer  += NAMED_RANGE_OPCODE + Lstring( cName, 16 ) + cRange
            ENDIF
         NEXT
         IF ! Empty( WKS_RANGE )
            buffer += PRINT_RANGE_OPCODE + WKS_RANGE
         ENDIF
         // Print format
         buffer += FORMAT_OPCODE + Chr( IF( WKS_PRINT == "F", 0, 1 ) )
         // Sort ranges
         IF ! Empty( WKS_SORT_RANGE )
            buffer += SORT_RANGE + WKS_SORT_RANGE
            // Sort key ranges
            IF ! Empty( WKS_SORTKEY1 )
               buffer += SORT_KEY_OP + WKS_SORTKEY1 + Chr( WKS_SORTKEY1_DIR )
            ENDIF
            IF ! Empty( WKS_SORTKEY2 )
               buffer += SORT_KEY2_OP + WKS_SORTKEY2 + Chr( WKS_SORTKEY2_DIR )
            ENDIF
         ENDIF
         // Protection
         buffer += PROTECT_OPCODE + Chr( IF( WKS_PROTECTION = "P", 1, 0 ) )
         // Footer
         IF WKS_VERSION <> "QP"
            buffer += FOOTER_OPCODE + Lstring( WKS_FOOTER, 242 )
         ELSE
            buffer += QP_FOOTER_OPCODE + Trim( WKS_FOOTER )
         ENDIF
         // Header
         IF WKS_VERSION <> "QP"
            buffer += HEADER_OPCODE + Lstring( WKS_HEADER, 242 )
         ELSE
            buffer += QP_HEADER_OPCODE + Trim( WKS_HEADER )
         ENDIF
         // Setup
         IF WKS_VERSION <> "QP"
            buffer += SETUP_OPCODE + Lstring( WKS_SETUP, 40 )
         ELSE
            buffer += QP_SETUP_OPCODE + Trim( WKS_SETUP )
         ENDIF
         // Margins
         IF WKS_VERSION <> "QP"
            buffer += MARGINS_OPCODE + I2Bin( WKS_LEFT ) + I2Bin( WKS_RIGHT ) + ;
                      I2Bin( WKS_LPP ) + I2Bin( WKS_TOP ) + I2Bin( WKS_BOTTOM )
         ELSE
            buffer += MARGINS_OPCODE + I2Bin( WKS_LEFT ) + I2Bin( WKS_RIGHT ) + ;
                      I2Bin( WKS_TOP ) + I2Bin( WKS_BOTTOM ) + I2Bin( WKS_LPP )
         ENDIF
         // Label format
         buffer += ALIGN_OPCODE + IF( WKS_ALIGNMENT == "C", Chr( 94 ), ;
                                      IF( WKS_ALIGNMENT == "R", Chr( 34 ), Chr( 39 ) ) )

         // Print borders
         IF ! Empty( WKS_BORDERS[ 1 ] ) .AND. ! Empty( WKS_BORDERS[ 2 ] )
            buffer += BORDERS_OPCODE + WKS_BORDERS[ 1 ] + WKS_BORDERS[ 2 ]
         ENDIF

      ELSE

******************************************************************************
* Lotus 3.x *
******************************************************************************
         // Special handling for Lotus 3.x spreadsheets
         // Calcset

         buffer += WK3_CALCSET + IF( WKS_CALCMODE == "A", I2Bin( 1 ), I2Bin( 0 ) ) + ;
                   IF( WKS_CALCORDER == "N", I2Bin( 0 ), IF( WKS_CALCORDER == "C", I2Bin( 1 ), I2Bin( 2 ) ) ) + ;
                   I2Bin( WKS_ITERATION )

         // Windowset
         buffer += WK3_WINDOWSET + Chr( 1 ) + Chr( 1 ) + Chr( 1 ) + Chr( 0 ) + I2Bin( 0 ) + ;
                   Chr( 8 ) + Chr( 20 ) + Chr( 72 ) + Chr( 4 ) + Chr( 4 ) + Chr( 0 ) + ;
                   I2Bin( 0 ) + Replicate( Chr( 0 ), 6 ) + I2Bin( 0 ) + Replicate( Chr( 0 ), 6 )

         // SheetCellPtr
         tmp := Lcell( WKS_FIRST_CELL, WKS_VERSION )
         buffer += WK3_SHEETCELLPTR + SubStr( tmp, 3, 1 ) + Chr( 0 ) + I2Bin( 0 ) + ;
                   SubStr( tmp, 1, 2 ) + SubStr( tmp, 4, 1 ) + Chr( 0 ) + I2Bin( 0 ) + I2Bin( 0 ) + ;
                   Chr( 0 ) + Chr( 0 ) + I2Bin( 0 )

         // Column widths
         IF Len( WKS_WIDTHS ) > 0
            ASort( WKS_WIDTHS )
            x   := Val( SubStr( WKS_WIDTHS[ 1 ], 1, 3 ) )
            tmp := ""
            FOR jj = 1 TO Len( WKS_WIDTHS )
               DO WHILE jj <= Len( WKS_WIDTHS ) .AND. Val( SubStr( WKS_WIDTHS[ jj ], 1, 3 ) ) = x
                  tmp   += SubStr( WKS_WIDTHS[ jj ], 4, 2 )
                  IF jj < Len( WKS_WIDTHS ) .AND. Val( SubStr( WKS_WIDTHS[ jj + 1 ], 1, 3 ) ) = x
                     jj ++
                  ELSE
                     EXIT
                  ENDIF
               ENDDO
               IF ! Empty( tmp )
                  buffer += WK3_WIDTH + Chr( x ) + Chr( 0 ) + I2Bin( 0 ) + tmp
                  tmp := ""
                  IF jj <= Len( WKS_WIDTHS )
                     tmp   := SubStr( WKS_WIDTHS[ jj ], 4, 2 )
                  ENDIF
               ENDIF
               x ++
            NEXT
         ENDIF

         // Range names - user ************************************************

         FOR jj = 1 TO Len( WKS_NAMED_RANGES )
            cName   := WKS_NAMED_RANGES[ jj ][ 1 ]
            cRange  := WKS_NAMED_RANGES[ jj ][ 2 ]
            buffer  += WK3_USERRANGE + I2Bin( 0 ) + Lstring( cName, 16 ) + cRange
         NEXT

         // Range names - system **********************************************

         IF ! Empty( WKS_SORT_RANGE )
            cName   := Chr( 25 ) + "DS0"
            buffer  += WK3_SYSTEMRANGE + I2Bin( 0 ) + Lstring( cName, 16 ) + WKS_SORT_RANGE
         ENDIF
         IF ! Empty( WKS_SORTKEY1 )
            cName   := Chr( 25 ) + Chr( 1 ) + "XS"
            buffer  += WK3_SYSTEMRANGE + I2Bin( 0 ) + Lstring( cName, 16 ) + WKS_SORTKEY1
         ENDIF
         IF ! Empty( WKS_SORTKEY2 )
            cName   := Chr( 25 ) + Chr( 2 ) + "XS"
            buffer  += WK3_SYSTEMRANGE + I2Bin( 0 ) + Lstring( cName, 16 ) + WKS_SORTKEY2
         ENDIF
         IF ! Empty( WKS_RANGE )
            cName   := Chr( 25 ) + Chr( 1 ) + "P3"
            buffer  += WK3_SYSTEMRANGE + I2Bin( 0 ) + Lstring( cName, 16 ) + WKS_RANGE
         ENDIF
         IF ! Empty( WKS_BORDERS[ 1 ] )
            cName   := Chr( 25 ) + Chr( 1 ) + "RLP"
            buffer  += WK3_SYSTEMRANGE + I2Bin( 0 ) + Lstring( cName, 16 ) + WKS_BORDERS[ 1 ]
         ENDIF
         IF ! Empty( WKS_BORDERS[ 2 ] )
            cName   := Chr( 25 ) + Chr( 1 ) + "RMP"
            buffer  += WK3_SYSTEMRANGE + I2Bin( 0 ) + Lstring( cName, 16 ) + WKS_BORDERS[ 2 ]
         ENDIF
         // Sort information **************************************************
         IF ! Empty( WKS_SORTKEY1 ) .OR. ! Empty( WKS_SORTKEY2 )
            buffer += WK3_SORTKEYDIR + Chr( IF( WKS_SORTKEY1_DIR = 0, 1, 0 ) ) + Chr( 0 ) + ;
                      Chr( IF( WKS_SORTKEY2_DIR = 0, 1, 0 ) ) + Chr( 0 )
         ENDIF

         // Print information *************************************************

         buffer += WK3_PRINTMAIN + Chr( 1 ) + Lstring( "", 16 ) + Lstring( "", 16 ) + ;
                   Lstring( "", 16 ) + Chr( 0 ) + Chr( IF( WKS_PRINT == "F", 1, 0 ) ) + ;
                   Replicate( Chr( 0 ), 19 ) + I2Bin( WKS_LEFT ) + ;
                   I2Bin( WKS_RIGHT ) + I2Bin( WKS_TOP ) + I2Bin( WKS_BOTTOM ) + I2Bin( 1 ) + ;
                   I2Bin( 1 ) + I2Bin( WKS_LPP ) + I2Bin( 0 )

         IF ! Empty( WKS_HEADER )
            buffer += WK3_PRINTSTRING + Chr( 1 ) + Chr( 0 ) + Lstring( WKS_HEADER, 242 )
         ENDIF
         IF ! Empty( WKS_FOOTER )
            buffer += WK3_PRINTSTRING + Chr( 1 ) + Chr( 1 ) + Lstring( WKS_FOOTER, 242 )
         ENDIF
         IF ! Empty( WKS_SETUP )
            buffer += WK3_PRINTSTRING + Chr( 1 ) + Chr( 2 ) + Lstring( WKS_SETUP, 242 )
         ENDIF

         // Format information
         IF ! Empty( WKS_FORMATS )
            // Load formats into the buffer string
            FOR jj = 1 TO Len( WKS_FORMATS )
               nSheet := Val( SubStr( WKS_FORMATS[ jj ], 1, 2 ) )
               nRow   := Val( SubStr( WKS_FORMATS[ jj ], 3, 4 ) )
               x      := Len( WKS_FORMATS[ jj ] )
               WHILE SubStr( WKS_FORMATS[ jj ], x, 1 ) == Chr( 192 )
                  x --
               ENDDO
               WKS_FORMATS[ jj ] := SubStr( WKS_FORMATS[ jj ], 1, x )
               outs := ""
               ct   :=  1
               x    :=  7
               WHILE x <= Len( WKS_FORMATS[ jj ] )
                  tmp  := SubStr( WKS_FORMATS[ jj ], x ++, 1 )
                  IF tmp == Chr( 224 )
                     tmp := Chr( 255 )
                  ENDIF
                  outs += Wk3_format( tmp, 1 )
               ENDDO
               buffer += WK3_FORMAT + I2Bin( Len( outs ) + 4 ) + Chr( nSheet ) + Chr( 0 ) + ;
                         I2Bin( nRow - 1 ) + outs
               FWrite( WKS_HANDLE, buffer, Len( buffer ) )
               buffer := ""
            NEXT
         ENDIF

         buffer += WK3_GBLFMT + Chr( 0 ) + Chr( 1 ) + I2Bin( 0 ) + ;
                   Chr( IF( WKS_PROTECTION = "P", 241, 113 ) ) + Chr( 0 ) + Chr( 0 ) + Chr( 0 ) + ;
                   IF( WKS_ALIGNMENT == "C", I2Bin( 94 ), IF( WKS_ALIGNMENT == "R", I2Bin( 34 ), I2Bin( 39 ) ) ) + ;
                   I2Bin( 0 )
      ENDIF
******************************************************************************
* Write the buffer created to the disk file *
******************************************************************************
      FWrite( WKS_HANDLE, buffer, Len( buffer ) )
      buffer := ""
      WKS_DATA_WRITTEN := .T.
   ENDIF
   RETURN NIL
*********************


* Function:    Wk3_format()
* Purpose:     Returns a Lotus 3.x format string
* Syntax:      wk3_format( cFormat,nCtr )
* Arguments:   cFormat      - Cell format
*              nCtr         - Number of times format repeats
* Returns:     cString      - Lotus 3.x format string
* Class:       Internal
*************************************
STATIC FUNCTION wk3_format( cForm, nCt )
   LOCAL temp
   IF nCt == 1
      temp := cForm + Chr( 0 ) + Chr( 0 ) + Chr( 0 )
   ELSE
      temp := cForm + Chr( 0 ) + Chr( 0 ) + Chr( 128 ) + Chr( nCt )
   ENDIF
   RETURN temp
*****************************************************************************
* End of Ldirty *
*****************************************************************************


* Function:    Makeformula()
* Purpose:     Takes the formula and converts it into a spreadsheet format
* Syntax:      makeformula( cFormula,cVersion,nRow,nColumn )
* Arguments:   cFormula
*              cVersion
*              nRow
*              nColumn
* Notes:       Makeformula() is used internally to convert a
*              character string containg an @ formula to its
*              internal format.  It is a STATIC function that
*              is only called by Lput() when Lput() is parsing a
*              character string containing a formula.
* Returns:     cToken      - data in spreadsheet tokenized format
* Class:       Internal
**************************************************
STATIC FUNCTION makeformula( cFormula, cVersion, nRow, nCol )
*
* Take <cFormula> and returns internal representation of that formula
*
*
* Formula := @xxx(   <param>   )
*
*
*                  ---------------------  Function code
*                 |      ---------------  Lotus 1-2-3 operation code
*                 |      |     ---------  Quattro/Quattro Pro operation code
*                 |      |   |
*                 |      |   |
*                 |      |   |
*                 |      |   |
*                 V      V   V
*
   STATIC form_ := { ;
                     { "SUM"  , 80, 29,  4, 99 }, ;         // Varying # of parameters
                     { "AVG"  , 81, 32,  5, 99 }, ;
                     { "COUNT", 82, 33,  0, 99 }, ;
                     { "MIN"  , 83, 34,  6, 99 }, ;
                     { "MAX"  , 84, 35,  7, 99 }, ;
                     { "STD"  , 88, 31, 12, 99 }, ;
                     { "VAR"  , 87, 30, 46, 99 }, ;
                     { "CHOSE", 48, 36, 16, 99 }, ;
                     { "NA"   , 31, 109, 10, 0 }, ;         // No parameter functions
                     { "ERR"  , 32, 110, 84, 0 }, ;
                     { "RAND" , 53, 111, 63, 0, "S" }, ;
                     { "PI"   , 38, 112, 19, 0 }, ;
                     { "FALSE", 51, 113, 35, 0 }, ;
                     { "TRUE" , 52, 114, 34, 0 }, ;
                     { "NOW"  , 55, 115, 74, 0, "S" }, ;
                     { "TODAY", 55, 118, 221, 0, "S" }, ;
                     { "ABS"  , 33, 124, 24, 1 }, ;         // One parameter functions
                     { "SQRT" , 35, 126, 20, 1 }, ;
                     { "LOG"  , 36, 127, 109, 1 }, ;
                     { "INT"  , 34, 125, 25, 1 }, ;
                     { "LN"   , 37, 128, 22, 1 }, ;
                     { "SIN"  , 39, 129, 15, 1 }, ;
                     { "COS"  , 40, 130, 16, 1 }, ;
                     { "TAN"  , 41, 131, 17, 1 }, ;
                     { "ATAN" , 43, 132, 18, 1 }, ;
                     { "ASIN" , 44, 133, 98, 1 }, ;
                     { "ACOS" , 45, 134, 99, 1 }, ;
                     { "EXP"  , 46, 135, 21, 1 }, ;
                     { "ISNA" , 49, 136,  2, 1 }, ;
                     { "ISERR", 50, 137,  3, 1 }, ;
                     { "DAY"  , 60, 138, 67, 1 }, ;
                     { "MONTH", 61, 139, 68, 1 }, ;
                     { "YEAR" , 62, 140, 69, 1 }, ;
                     { "ISNO" , 68, 141, 128, 1 }, ;
                     { "ISLAB", 69, 142, 127, 1 }, ;
                     { "COLS" , 99, 145, 77, 1 }, ;
                     { "ROWS" , 100, 146, 76, 1 }, ;
                     { "CHAR" , 74, 147, 111, 1 }, ;
                     { "CODE" , 75, 148, 121, 1 }, ;
                     { "LEN"  , 70, 149, 211, 1 }, ;
                     { "LOWER", 103, 150, 112, 1 }, ;
                     { "PROPR", 107, 152, 114, 1 }, ;
                     { "TRIM" , 109, 154, 118, 1 }, ;
                     { "UPPER", 102, 155, 113, 1 }, ;
                     { "VALUE", 71, 156, 33, 1 }, ;
                     { "DVAL" , 77, 157, 140, 1 }, ;
                     { "TVAL" , 78, 158, 141, 1 }, ;
                     { "HOURS", 65, 159, 71, 1 }, ;
                     { "MINS" , 66, 160, 72, 1 }, ;
                     { "SECS" , 67, 161, 73, 1 }, ;
                     { "DEG"  ,  0, 163, - 1, 1 }, ;
                     { "RAD"  ,  0, 164, - 1, 1 }, ;
                     { "HEX2N",  0, 165, - 1, 1 }, ;
                     { "NUM2H",  0, 166, - 1, 1 }, ;
                     { "FILE" ,  0, 167, - 1, 1 }, ;
                     { "CLEAN",  0, 168, - 1, 1 }, ;
                     { "MOD"  , 47, 187, - 1, 2 }, ;         // Two parameter functions
                     { "ATAN2", 42, 188, 18, 2 }, ;
                     { "ROUND", 63, 189, 27, 2 }, ;
                     { "NPV"  , 86, 190, 11, 2 } }


   LOCAL retval := "", x, lookfor, params := NIL, needs := 0, nFormcode, cPar, vctr := 0
   LOCAL aQPro := { "", "", "" }
   LOCAL nOffset := 2
   IF Len( form_ ) < 70
      AAdd( form_, {  "IRR"  , 89, 191, 62, 2 } )
      AAdd( form_, {  "CELL" , 108, 192, 94, 2 } )
      AAdd( form_, {  "EXACT", 113, 193, 117, 2 } )
      AAdd( form_, {  "LEFT" , 104, 194, 115, 2 } )
      AAdd( form_, {  "RIGHT", 105, 196, 116, 2 } )
      AAdd( form_, {  "STRNG", 72, 197, - 1, 2 } )
      AAdd( form_, {  "DATE" , 54, 206, 65, 3 } )         // Three parameter functions
      AAdd( form_, {  "PMT"  , 56, 207, 59, 3 } )
      AAdd( form_, {  "PV"   , 57, 208, 56, 3 } )
      AAdd( form_, {  "FV"   , 58, 209, 57, 3 } )
      AAdd( form_, {  "IF"   , 59, 59,  0, 3 } )
      AAdd( form_, {  "VLOOK", 85, 211, 102, 3 } )
      AAdd( form_, {  "HLOOK", 90, 212, 101, 3 } )
      AAdd( form_, {  "INDEX", 98, 213, - 1, 3 } )
      AAdd( form_, {  "FIND" , 76, 214, 124, 3 } )
      AAdd( form_, {  "MID"  , 73, 215, - 1, 3 } )
      AAdd( form_, {  "TIME" , 64, 216, 66, 3 } )
      AAdd( form_, {  "CTERM", 118, 217, - 1, 3 } )
      AAdd( form_, {  "RATE" , 116, 218, 60, 3 } )
      AAdd( form_, {  "SLN"  , 119, 219, 142, 3 } )
      AAdd( form_, {  "DSTD" , 97, 220, 45, 3 } )
      AAdd( form_, {  "DVAR" , 96, 221, 47, 3 } )
      AAdd( form_, {  "DSUM" , 91, 222, 41, 3 } )
      AAdd( form_, {  "DAVG" , 92, 223, 43, 3 } )
      AAdd( form_, {  "DCNT" , 93, 224, 40, 3 } )
      AAdd( form_, {  "DMIN" , 94, 225, 43, 3 } )
      AAdd( form_, {  "DMAX" , 95, 226, 44, 3 } )
      AAdd( form_, {  "TERM" , 117, 227, - 1, 3 } )
      AAdd( form_, {  "REPL" , 106, 232, 119, 4 } )         // Four parameter functions
      AAdd( form_, {  "DDB"  , 121, 233, 144, 4 } )
      AAdd( form_, {  "SYD"  , 120, 234, 143, 4 } )
   ENDIF

   cVersion := IF( cVersion = NIL, "L2", cVersion )
   IF cVersion == "QU"
      nOffset := 3
   ELSEIF SubStr( cVersion, 1, 1 ) == "E"
      nOffset := 4
   ENDIF


   IF cFormula <> NIL .AND. ValType( cFormula ) == "C"
      // Remove leading @ from formula string
      cFormula := IF( SubStr( cFormula, 1, 1 ) == "@", SubStr( cFormula, 2 ), cFormula )
      cFormula := IF( SubStr( cFormula, 1, 1 ) == "=" .AND. SubStr( cVersion, 1, 1 ) == "E", SubStr( cFormula, 2 ), cFormula )
      x := At( "(", cFormula )
      IF x > 0
         cFormula := SubStr( cFormula, 1, Len( cFormula ) - 1 )   // Remove final )
         lookfor := SubStr( cFormula, 1, x - 1 )                // Extract formula
         params  := SubStr( cFormula, x + 1 )                  // and parameters
      ELSE
         lookfor := cFormula                              // No parameters
      ENDIF
      x := AScan( form_, { | y | y[ 1 ] == Upper( lookfor ) } )     // Look for match
      IF x > 0
         // Quattro or Lotus formula number ?
         nFormcode := form_[ x, nOffset ]
         needs     := form_[ x, 5 ]                          // Number of parameters
         IF needs == 0                                    // No parameters needed
            IF cVersion <> "QP"
               IF SubStr( cVersion, 1, 1 ) == "E"
                  retval   := Chr( 65 ) + IF( cVersion == "E4", I2Bin( nFormcode ), ;
                                              Chr( nFormcode ) )
               ELSE
                  retval   := Chr( nFormcode )
               ENDIF
            ELSE
               aQPro[ QP_FORMULAS ] := Chr( nFormcode )
            ENDIF
         ELSEIF params <> NIL                             // Parameters found
            WHILE ! Empty( params )
               x := At( ",", params )
               vctr ++
               IF x > 0
                  cPar  := SubStr( params, 1, x - 1 )
                  params := SubStr( params, x + 1 )
               ELSE
                  cPar  := params
                  params := ""
               ENDIF
               IF Val( cPar ) == 0 .OR. ( nFormcode == 59 .AND. vctr = 1 )
                  IF cVersion <> "QP"
                     retval += To_123( cPar, cVersion, nRow, nCol, vctr = 1 .AND. nFormcode = 59 )
                  ELSE
                     IF ".." $ cPar
                        aQPro[ QP_FORMULAS ] += To_123( "RANGEMARK", "QP", nRow, nCol )
                        aQPro[ QP_RANGES ] += To_123( cPar, cVersion, nRow, nCol )
                     ELSEIF Len( cPar ) > 1
                        aQPro[ QP_FORMULAS ] += To_123( "CELL_ADDR", "QP", nRow, nCol )
                        aQPro[ QP_CELLS ]  += To_123( cPar, cVersion, nRow, nCol )
                     ELSE
                        aQPro[ QP_FORMULAS ]  += To_123( cPar, cVersion, nRow, nCol, vctr = 1 .AND. nFormcode = 59 )
                     ENDIF
                  ENDIF
               ELSE
                  IF cVersion <> "QP"
                     retval += To_123( Val( cPar ), cVersion, nRow, nCol )
                  ELSE
                     aQPro[ QP_FORMULAS ] += To_123( Val( cPar ), cVersion, nRow, nCol )
                  ENDIF
               ENDIF
            ENDDO
            IF ! Empty( retval ) .OR. cVersion == "QP"
               IF cVersion <> "QP"
                  IF SubStr( cVersion, 1, 1 ) == "E"
                     IF cVersion <> "E4"
                        retval += IF( needs == 99, Chr( 66 ) + Chr( vctr ), Chr( 65 ) ) + Chr( nFormcode )
                     ELSE
                        retval += IF( needs == 99, Chr( 66 ) + Chr( vctr ), Chr( 65 ) ) + I2Bin( nFormcode )
                     ENDIF
                  ELSE
                     retval += Chr( nFormcode ) + IF( needs == 99, Chr( vctr ), "" )
                  ENDIF
               ELSE
                  aQPro[ QP_FORMULAS ] += Chr( nFormcode ) + IF( needs == 99, Chr( vctr ), "" )
               ENDIF
            ENDIF
         ENDIF
      ENDIF
   ENDIF
   RETURN IF( cVersion <> "QP", retval, aQPro )
***************************


* Function:    To_123()
* Purpose:     Takes the data and converts it into a spreadsheet token
* Syntax:      To_123( xData,cVersion,nRow,nColumn )
* Arguments:   xData
*              cVersion
*              nRow
*              nColumn
* Notes:       This function is used by Lput() to convert data of
*              any type to the internal format used by
*              spreadsheets for storage.  It should never be
*              called directly.
*
*              Columns
*              -------
*                 2        Lotus 2.x
*                 3        Quattro
*                 4        Excel
*                 5        Lotus 3.x
*
* Returns:     cToken      - data in spreadsheet tokenized format
* Class:       Internal
**************************************************
STATIC FUNCTION to_123( xData, cVersion, nRow, nCol, lIsExp )
   STATIC markers_ := { ;
                        { "ENDMARK"  , 3, 25, 0, 3 }, ;
                        { "RANGEMARK", 2, 17, 37, 2 }, ;
                        { "INTEGER"  , 5, 18, 30, 5 }, ;
                        { "REAL"     , 0, 19, 31, 0 }, ;
                        { "CELL_ADDR", 1, 20, 68, 1 }, ;
                        { "LABEL"    , 6, 21, 0, 6 }, ;
                        { "+"        , 9,  3, 3, 15 }, ;
                        { "-"        , 10,  4, 4, 16 }, ;
                        { "!"        , 8,  4, 4, 14 }, ;
                        { "*"        , 11,  5, 5, 17 }, ;
                        { "/"        , 12,  6, 6, 18 }, ;
                        { "<="       , 16, 16, 10, 22 }, ;
                        { ">="       , 17, 17, 12, 23 }, ;
                        { "<>"       , 15, 15, 15, 21 }, ;
                        { "^"        , 13,  8, 7, 19 }, ;
                        { "="        , 14, 14, 11, 20 }, ;
                        { "<"        , 18, 18, 9, 24 }, ;
                        { ">"        , 19, 19, 13, 25 }, ;
                        { ")"        , 04, 23, 21, 04 }  }

   LOCAL retval := "", nElt, x := 0, tmp
   LOCAL r1, c1, r2, c2, tmp1, tmp2, s1 := 0, s2 := 0
   LOCAL nSize := 1, arr_ := { xData }
   LOCAL xx, zz, p_
   LOCAL k, temp_out, y
   //LOCAL nTest,

   cVersion := IF( cVersion == NIL, "L2", SubStr( cVersion, 1, 2 ) )
   nElt     := IF( cVersion == "QU", 3, ;
                   IF( SubStr( cVersion, 1, 1 ) == "E", 4, IF( cVersion = "L2", 2, 5 ) ) )
   lIsExp   := IF( lIsExp = NIL, .F., lIsExp )
   IF lIsExp
      FOR zz := 12 TO 19
         xx := At( markers_[ zz, 1 ], xData )
         IF xx > 0
            ASize( arr_, 3 )
            arr_[ 1 ] := SubStr( xData, 1, xx - 1 )
            arr_[ 2 ] := SubStr( xData, xx + Len( markers_[ zz, 1 ] ) )
            arr_[ 3 ] := markers_[ zz, 1 ]
            IF Val( arr_[ 1 ] ) > 0
               arr_[ 1 ] := Val( arr_[ 1 ] )
            ENDIF
            IF Val( arr_[ 2 ] ) > 0
               arr_[ 2 ] := Val( arr_[ 2 ] )
            ENDIF
            nSize   := 3
            zz      := 99
         ENDIF
      NEXT
   ENDIF


   FOR xx := 1 TO nSize
      xData := arr_[ xx ]
      IF ValType( xData ) = "C" .AND. Replicate( "0", Len( xData ) ) = xData
         xData := 0
      ENDIF
      IF ValType( xData ) == "C"
         DO CASE
         CASE Len( xData ) <= 1 .AND. ;
            ( x := AScan( markers_, { | y | y[ 1 ] == xData }, 7, 13 ) ) > 0
            IF cVersion $ "L3|L4"
               retval += Chr( markers_[ x, nElt ] )
            ELSE
               retval += Chr( markers_[ x, nElt ] )
            ENDIF

         CASE Len( xData ) == 2 .AND. ;
            ( x := AScan( markers_, { | y | y[ 1 ] == xData }, 12, 3 ) ) > 0
            IF cVersion $ "L3|L4"
               retval += Chr( markers_[ x, nElt ] )
            ELSE
               retval += Chr( markers_[ x, nElt ] )
            ENDIF

         CASE ".." $ xData .OR. ( ":" $ xData .AND. SubStr( cVersion, 1, 1 ) == "E" )
            tmp := Lmakerange( xData, cVersion )   // Convert range to internal format
            IF cVersion $ "L3|L4"
               r1  := Bin2W( SubStr( tmp, 1, 2 ) )
               s1  := Asc( SubStr( tmp, 3, 1 ) )
               c1  := Asc( SubStr( tmp, 4, 1 ) )
               r2  := Bin2W( SubStr( tmp, 5, 2 ) )
               s2  := Asc( SubStr( tmp, 7, 1 ) )
               c2  := Asc( SubStr( tmp, 8, 1 ) )
            ELSE
               r1  := Bin2W( SubStr( tmp, 1, 2 ) )
               c1  := Bin2W( SubStr( tmp, 3, 2 ) )
               r2  := Bin2W( SubStr( tmp, 5, 2 ) )
               c2  := Bin2W( SubStr( tmp, 7, 2 ) )
            ENDIF
            IF cVersion $ "L3|L4"
               // Lotus 3 range markes
               IF "$" $ xData
                  retval += Chr( markers_[ 2, nElt ] ) + Chr( 0 ) + ;
                            I2Bin( r1 ) + Chr( s1 ) + Chr( c1 ) + I2Bin( r2 ) + Chr( s2 ) + Chr( c2 )
               ELSE
                  retval += Chr( markers_[ 2, nElt ] ) + Chr( 255 ) + ;
                            I2Bin( r1 ) + Chr( s1 ) + Chr( c1 ) + I2Bin( r2 ) + Chr( s2 ) + Chr( c2 )
               ENDIF
            ELSEIF cVersion == "QP"
               IF "$" $ xData
                  retval += I2Bin( r1 ) + I2Bin( c1 ) + I2Bin( r2 ) + I2Bin( c2 )
               ELSE
                  retval += I2Bin( Lrelative( r1, nRow ) ) + I2Bin( Lrelative( c1, nCol ) ) + ;
                            I2Bin( Lrelative( r2, nRow ) ) + I2Bin( Lrelative( c2, nCol ) )
               ENDIF
            ELSEIF SubStr( cVersion, 1, 1 ) == "E"
               IF "$" $ xData
                  retval += Chr( markers_[ 2, nElt ] ) + I2Bin( r1 ) + I2Bin( r2 ) + Chr( c1 ) + Chr( c2 )
               ELSE
                  tmp1 := I2Bin( r1 )
                  tmp2 := I2Bin( r2 )
                  bit( tmp1, 10, .T. )
                  bit( tmp1, 9, .T. )
                  bit( tmp2, 10, .T. )
                  bit( tmp2, 9, .T. )
                  retval := Chr( markers_[ 2, nElt ] ) + tmp1 + tmp2 + Chr( c1 ) + Chr( c2 )
               ENDIF
            ELSE
               IF "$" $ xData
                  retval += Chr( markers_[ 2, nElt ] ) + ;
                            I2Bin( r1 ) + I2Bin( c1 ) + I2Bin( r2 ) + I2Bin( c2 )
               ELSE
                  retval += Chr( markers_[ 2, nElt ] ) + ;
                            I2Bin( Lrelative( r1, nRow ) ) + I2Bin( Lrelative( c1, nCol ) ) + ;
                            I2Bin( Lrelative( r2, nRow ) ) + I2Bin( Lrelative( c2, nCol ) )
               ENDIF
            ENDIF
         OTHERWISE
            x := AScan( markers_, { | y | Upper( y[ 1 ] ) == Upper( xData ) } )
            DO CASE
            CASE x > 0
               retval += Chr( markers_[ x, nElt ] )
            CASE Left( xData, 1 ) $ "@="      // Some formula
               retval += makeformula( xData, cVersion, nRow, nCol )
            CASE Left( xData, 1 ) $ "+-"       // Some expression
               p_       := Parse( xData )
               temp_out := ""
               IF ! Empty( p_ )
                  IF cVersion <> "QP" .AND. Left( cVersion, 1 ) <> "E"
                     FOR k = Len( p_ ) TO 1 STEP - 1
                        IF SubStr( p_[ k ], 1, 1 ) == "@"
                           temp_out += makeformula( p_[ k ], cVersion, nRow, nCol )
                        ELSEIF Val( p_[ k ] ) == 0
                           y := Len( p_[ k ] )
                           IF p_[ k ] <> Replicate( "0", y )
                              temp_out += to_123( p_[ k ], cVersion, nRow, nCol )
                           ELSE
                              temp_out += to_123( Val( p_[ k ] ), cVersion, nRow, nCol )
                           ENDIF
                        ELSE
                           temp_out += to_123( Val( p_[ k ] ), cVersion, nRow, nCol )
                        ENDIF
                     NEXT
                  ELSEIF Left( cVersion, 1 ) == "E"
                     FOR k = Len( p_ ) TO 1 STEP - 1
                        IF SubStr( p_[ k ], 1, 1 ) == "="
                           temp_out += makeformula( p_[ k ], cVersion, nRow, nCol )
                        ELSEIF Val( p_[ k ] ) == 0
                           y := Len( p_[ k ] )
                           IF p_[ k ] <> Replicate( "0", y )
                              temp_out += to_123( p_[ k ], cVersion, nRow, nCol )
                           ELSE
                              temp_out += to_123( Val( p_[ k ] ), cVersion, nRow, nCol )
                           ENDIF
                        ELSE
                           temp_out += to_123( Val( p_[ k ] ), cVersion, nRow, nCol )
                        ENDIF
                     NEXT
                  ENDIF
               ENDIF
               retval += temp_out
            OTHERWISE                    // Assumed to be a cell address
               tmp  := Lcell( xData, cVersion )
               r1   := Bin2W( SubStr( tmp, 1, 2 ) )
               IF cVersion $ "L3|L4"
                  c1   := Asc( SubStr( tmp, 4, 1 ) )
                  s1   := Asc( SubStr( tmp, 3, 1 ) )
               ELSE
                  c1   := Bin2W( SubStr( tmp, 3, 2 ) )
               ENDIF
               IF cVersion $ "L3|L4"
                  IF "$" $ xData
                     retval += Chr( markers_[ 5, nElt ] ) + Chr( 0 ) + ;
                               I2Bin( r1 ) + Chr( s1 ) + Chr( c1 )
                  ELSE
                     retval += Chr( markers_[ 5, nElt ] ) + Chr( 07 ) + ;
                               I2Bin( r1 ) + Chr( s1 ) + Chr( c1 )
                  ENDIF
               ELSEIF SubStr( cVersion, 1, 1 ) == "E"
                  IF "$" $ xData
                     retval += Chr( markers_[ 5, nElt ] ) + I2Bin( r1 ) + Chr( c1 )
                  ELSE
                     tmp1   := I2Bin( r1 )
                     bit( tmp1, 10, .T. )
                     bit( tmp1, 9, .T. )
                     retval := Chr( markers_[ 5, nElt ] ) + tmp1 + Chr( c1 )
                  ENDIF
               ELSEIF cVersion <> "QP"
                  IF "$" $ xData
                     retval += Chr( markers_[ 5, nElt ] ) + ;
                               I2Bin( r1 ) + I2Bin( c1 )
                  ELSE
                     retval += Chr( markers_[ 5, nElt ] ) + ;
                               I2Bin( Lrelative( r1, nRow ) ) + ;
                               I2Bin( Lrelative( c1, nCol ) )
                  ENDIF
               ELSE
                  IF "$" $ xData
                     retval += I2Bin( r1 ) + I2Bin( c1 )
                  ELSE
                     retval += I2Bin( Lrelative( r1, nRow ) ) + ;
                               I2Bin( Lrelative( c1, nCol ) )
                  ENDIF
               ENDIF
            ENDCASE
         ENDCASE
      ELSEIF ValType( xData ) == "N"
         tmp := Str( xData )
         x   := IF( At( ".", tmp ) == 0, 0, Len( tmp ) - At( ".", tmp ) )
         IF ! ( cVersion $ "L3|L4" )
            IF x > 0
               retval += Chr( markers_[ 4, nElt ] ) + ieereal( xData )
            ELSE
               retval += Chr( markers_[ 3, nElt ] ) + I2Bin( xData )
            ENDIF
         ELSE
            IF xData >= - 16384 .AND. xData <= 16383 .AND. x == 0
               retval += Chr( markers_[ 3, nElt ] ) + shortnum( xData )
            ELSE
               retval += Chr( markers_[ 4, nElt ] ) + treal( xData )
            ENDIF
         ENDIF
      ENDIF
   NEXT
   RETURN retval
*************************************



* Function:    Parse()
* Purpose:     Takes an expression and returns a RPN stack in an array
* Syntax:      Parse( cExpression )
* Arguments:   cExpression - Formula to convert to RPN
* Returns:     aPolish    - Reverse polish notation array of tokens
* Notes:       Parse() is used to break a formula string down
*              into an array of reverse polish notation.  It is
*              used internally by Lput() and probably should not
*              be called directly by your application.
* Class:       Internal
**************************************************
STATIC FUNCTION parse( cExp )           // Thanks Mike Taylor
   #define EOT ''         // end-of-expression symbol
   #define NEGATE '!'      // unary '-' negation stack symbol
   #define IS_NUM( x )      ( x $ "0123456789" )
   #define IS_OP( x )       ( x $ "()+-*/^" + EOT )
   #define OPERAND  0      // parsing engine expects an operand mode
   #define OPERATOR 1      // parsing enging expects an operator mode
   #define ERR_PARENS 1    // mis-matched parens
   #define ERR_SYNTAX 2    // next token did not match state

   LOCAL cToken      := ""
   LOCAL cWorkingExp := StrTran( cExp, " ", "" ) + EOT   // remove ALL spaces
   LOCAL aOperators  := { }
   LOCAL aRevPolish  := { }
   LOCAL nState      := OPERAND
   LOCAL lError      := .F., lECode := 0, c := ""
   //LOCAL i := 0
   WHILE ! Empty( cWorkingExp )
      cToken := GetToken( @cWorkingExp, nState )
      IF ( ! Empty( cToken ) )
         DO CASE
         CASE cToken == EOT                    // last item in expression
            WHILE .T.
               c := Pop( aOperators )            // clear out remaining operators
               IF ( c != NIL )
                  Push( aRevPolish, c )
               ELSE
                  EXIT
               ENDIF
            ENDDO
         CASE cToken == "-" .AND. nState == OPERAND      // unary -
            Push( aOperators, NEGATE )
            nState := OPERAND
         CASE cToken == "+" .AND. nState == OPERAND      // unary +
            nState := OPERAND                            //  ignore it
         CASE cToken == "("                              // left paren
            Push( aOperators, cToken )                     //  push
            nState := OPERAND
         CASE cToken == ")"                              // if token == right paren
            WHILE .T.                                    //  pop operators until a
               c := Pop( aOperators )                      //  left paren is found
               DO CASE
               CASE c == NIL
                  lError := .T.
                  lECode := ERR_PARENS
                  EXIT
               CASE c == "("
                  EXIT
               OTHERWISE
                  Push( aRevPolish, c )
               ENDCASE
            ENDDO
            Push( aRevPolish, cToken )        // push the right paren
            nState := OPERATOR
         OTHERWISE
            IF ( nState == OPERAND )
               Push( aRevPolish, cToken )
               nState := OPERATOR
            ELSE
               OpCopy( aOperators, aRevPolish, cToken )
               nState := OPERAND
            ENDIF
         ENDCASE
      ENDIF
   ENDDO
   RETURN ( aRevPolish )
*******************


* Function:    Opcopy()
* Purpose:     Place token on the appropiate stack
* Syntax:      Opcopy( aOperators, aPolish, cToken )
* Arguments:   aOperators - Operations stack
*              aPolish    - Data stack
*              cToken     - Token being processed
* Returns:     NIL
* Notes:       Takes cToken and if determines which stack to push it onto.
*              If the operators stack has a token that is of lower precedence,
*              then they are switched otherwise it is pushed onto the stack.
* Class:       Internal
**************************************************
STATIC FUNCTION opcopy( aOperators, aRevPolish, cToken )
   LOCAL aTable := { "+", "-", "*", "/", "^" }
   LOCAL aPVal  := {   1,   1,   2,   2,   3 }
   LOCAL nTokenPrecedence := 0         // current token's precedence value
   LOCAL nTopPrecedence   := 0         // top operators stack's precedence
   //LOCAL cTemp            := ""

   // are there any operators on the stack?

   IF ( Len( aOperators ) == 0 )
      Push( aOperators, cToken )        // no, push current token
   ELSE
      // scan operator table for current token
      IF ( ( nTokenPrecedence := AScan( aTable, cToken ) ) != 0 )
         // get current tokens precedence value
         nTokenPrecedence := aPVal[ nTokenPrecedence ]
         nTopPrecedence   := AScan( aTable, Top( aOperators ) )
         nTopPrecedence   := IF( nTopPrecedence != 0, aPVal[ nTopPrecedence ], 0 )
         // check to see if current token's precedence value is
         //  less than or equal to the top token in the stack
         IF ( nTokenPrecedence <= nTopPrecedence )
            // if it is, then switch the top token with the current token
            Push( aRevPolish, Pop( aOperators ) )
            Push( aOperators, cToken )
         ELSE
            Push( aOperators, cToken )
         ENDIF
      ELSE
         Push( aOperators, cToken )
      ENDIF
   ENDIF
   RETURN NIL
*********************************


* Function:    Gettoken()
* Purpose:     Extracts from the expression, the next available token
* Syntax:      Gettoken( cExpression,nState )
* Arguments:   cExpression - Expression being processed, must be passed
*                            by reference
*              nState      - 0 = Operands    1 = Operators
* Notes:       Gettoken() is used internally by the PARSE()
*              function.  It is used to extract tokens from the
*              formula string to place on the stack.  It is a
*              STATIC function in the WRITEWKS.PRG and should
*              never be called directly by your application.
* Returns:     cToken      - Extracted token from expression
* Class:       Internal
**************************************
STATIC FUNCTION gettoken( cExp, nState )
   LOCAL cCurrChar := "", cToken := "", lDone := .F., p := 0

   WHILE ! ( lDone )
      cCurrChar := SubStr( cExp, 1, 1 )
      IF ( nState == OPERATOR )
         IF ( IS_OP( cCurrChar ) )
            cToken += cCurrChar
         ENDIF
         lDone := .T.
         cExp  := SubStr( cExp, 2 )
      ELSE
         IF ( cCurrChar == "@" ) .OR. ( cCurrChar == "=" )
            p := At( ")", cExp )
            IF ( p > 0 )
               cToken += SubStr( cExp, 1, p )
               cExp := SubStr( cExp, p + 1 )
               lDone := .T.
            ELSE
               cToken += cCurrChar
            ENDIF
         ELSE
            IF ( cCurrChar $ "()-+" + EOT )
               IF ( Empty( cToken ) )
                  cToken += cCurrChar
                  cExp   := SubStr( cExp, 2 )
               ENDIF
               lDone := .T.
            ELSE
               IF ( ! IS_OP( cCurrChar ) )
                  cToken += cCurrChar
                  cExp   := SubStr( cExp, 2 )
               ELSE
                  lDone := .T.
               ENDIF
            ENDIF
         ENDIF
      ENDIF
   ENDDO
   RETURN ( cToken )
**********************


* Function:    Push()
* Purpose:     Pushes the requested element onto the requested stack
* Syntax:      Push( aStack,xElement )
* Arguments:   aStack   - the stack being processed
*              xElement - the data being added to the stack
* Notes:       Pushes the element onto either the operator or
*              operand stack.  This function is a STATIC function
*              used only by the Parse() function.  You should
*              never need to call Push() directly.
* Returns:     NIL
* Class:       Internal
**************************************************

STATIC FUNCTION push( aStack, xElement )
   IF ( xElement != NIL )

      IF ( ( Len( aStack ) == 0 ) .OR. ( aStack[ Len( aStack ) ] != NIL ) )
         ASize( aStack, Len( aStack ) + 1 )
      ENDIF
      AIns( aStack, 1 )
      aStack[ 1 ] := xElement

   ENDIF
   RETURN NIL
*****************************************************************************
* End of Push *
*****************************************************************************



* Function:    Pop()
* Purpose:     Pops the top element from the stack
* Syntax:      Pop( aStack )
* Arguments:   aStack - the stack being processed
* Returns:     xData  - First item on the stack, NIL if stack is empty
* Notes:       The Pop() function removes the top element from
*              the stack and adjusts the stack size.  If there
*              are no elements in the stack, the function returns
*              NIL.  It is a STATIC function used by the Parse()
*              function and should never need to be called
*              directly by your application.
* Class:       Internal
**************************************************

STATIC FUNCTION pop( aStack )
   LOCAL xElement := NIL
   IF ( Len( aStack ) > 0 )
      xElement := aStack[ 1 ]
      ADel( aStack, 1 )
      ASize( aStack, Len( aStack ) - 1 )
   ENDIF
   RETURN ( xElement )
*****************************************************************************
* End of Pop *
*****************************************************************************



* Function:    Top()
* Purpose:     Returns the top element from the stack without removing it
* Syntax:      Top( aStack )
* Arguments:   aStack - the stack being processed
* Returns:     xData  - First item on the stack, NIL if stack is empty
* Notes:       Top() reads the operator stack and returns the
*              value of the first element, without removing it.
*              It is a STATIC function used by Lput() to write
*              formulas to the disk.  You should never need to
*              call Lput() directly.
* Class:       Internal
**************************************************

STATIC FUNCTION top( aStack )
   RETURN ( IF( Len( aStack ) == 0, NIL, aStack[ 1 ] ) )
*****************************************************************************
* End of Top *
*****************************************************************************
